import enum
import json
from typing import Set, Dict, Optional, Iterable, Callable, Union

from gi.repository import GLib  # type: ignore
from pydbus import SessionBus, SystemBus
from pydbus.bus import Bus
from defusedxml import ElementTree

from platypush.context import get_bus
from platypush.message import Message
from platypush.message.event import Event
from platypush.message.event.dbus import DbusSignalEvent
from platypush.message.request import Request
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.dbus import DbusSignalSchema
from platypush.utils import run


_default_service_name = 'org.platypush.Bus'
_default_service_path = '/'
_default_interface_name = 'org.platypush.Bus'


class BusType(enum.Enum):
    SYSTEM = 'system'
    SESSION = 'session'


class DBusService:
    """
    <node>
      <interface name="org.platypush.Bus">
        <method name="Post">
          <arg type="s" name="msg" direction="in"/>
          <arg type="s" name="response" direction="out"/>
        </method>
      </interface>
    </node>
    """

    @classmethod
    def _parse_msg(cls, msg: Union[str, dict]) -> dict:
        return Message.build(json.loads(json.dumps(msg)))

    def Post(self, msg: dict):
        """
        This method accepts a message as a JSON object
        (either representing a valid request or an event) and either
        executes it (request) or forwards it to the application bus (event).

        :param msg: Request or event, as a dictionary.
        :return: The return value of the request, or 0 if the message is an event.
        """
        msg = self._parse_msg(msg)
        if isinstance(msg, Request):
            ret = run(msg.action, **msg.args)
            if ret is None:
                ret = ''  # DBus doesn't like None return types

            return ret

        if isinstance(msg, Event):
            get_bus().post(msg)
            return 0


class DbusPlugin(RunnablePlugin):
    """
    Plugin to interact with DBus.

    This plugin can be used for the following tasks:

        * It can expose a D-Bus interface that other applications can use to push messages
          to Platypush (either action requests or events) serialized in JSON format.
          You can disable this listener by setting ``service_name`` to ``null`` in your
          configuration. If the D-Bus Platypush interface is enabled then you can push
          Platypush events and requests in JSON format from another application or script
          by specifying:

              * The D-Bus service (default: ``org.platypush.Bus``)
              * The D-Bus interface (default: ``org.platypush.Bus``)
              * The D-Bus path (default: ``/``)
              * The D-Bus method (``Post``)
              * The Platypush JSON payload (first argument of the request). Format:
                ``{"type": "request", "action": "module.action", "args": {...}}``

        * It can subscribe to multiple D-Bus signals, and it triggers a ``DbusSignalEvent``
          when an event is received (signal filters should be specified in the ``signals``
          configuration).

        * It can be used to query and inspect D-Bus objects through the :meth:`.query` method.

        * It can be used to execute methods exponsed by D-Bus objects through the
          :meth:`.execute` method.

    """

    def __init__(
        self,
        signals: Optional[Iterable[dict]] = None,
        service_name: Optional[str] = _default_service_name,
        service_path: Optional[str] = _default_service_path,
        **kwargs,
    ):
        """
        :param signals: Specify this if you want to subscribe to specific DBus
            signals. Structure:

            .. schema:: dbus.DbusSignalSchema(many=True)

            For example, to subscribe to all the messages on the session bus:

            .. code-block:: yaml

                dbus:
                    signals:
                        - bus: session

        :param service_name: Name of the D-Bus service where Platypush will listen
            for new messages (requests and events). Set to null if you want to disable
            message execution over D-Bus for Platypush (default: ``org.platypush.Bus``).

        :param service_path: The path of the D-Bus message listener. Set to null
            if you want to disable message execution over D-Bus for Platypush
            (default: ``/``).
        """
        super().__init__(**kwargs)
        self._system_bus = SystemBus()
        self._session_bus = SessionBus()
        self._loop = None
        self._signals = DbusSignalSchema().load(signals or [], many=True)
        self._signal_handlers = [
            self._get_signal_handler(**signal) for signal in self._signals
        ]

        self.service_name = service_name
        self.service_path = service_path

    @staticmethod
    def _get_signal_handler(bus: str, **_) -> Callable:
        def handler(sender, path, interface, signal, params):
            get_bus().post(
                DbusSignalEvent(
                    bus=bus,
                    signal=signal,
                    path=path,
                    interface=interface,
                    sender=sender,
                    params=params,
                )
            )

        return handler

    def _get_bus(self, bus_type: Union[str, BusType]) -> Bus:
        if isinstance(bus_type, str):
            bus_type = BusType(bus_type.lower())
        return self._system_bus if bus_type == BusType.SYSTEM else self._session_bus

    def _init_signal_listeners(self):
        for i, signal in enumerate(self._signals):
            handler = self._signal_handlers[i]
            bus = self._get_bus(signal['bus'])
            bus.subscribe(
                signal_fired=handler,
                signal=signal.get('signal'),
                sender=signal.get('sender'),
                object=signal.get('path'),
                iface=signal.get('interface'),
            )

    def _init_service(self):
        if not (self.service_name and self.service_path):
            return

        self._session_bus.publish(
            self.service_name,
            ('/', DBusService()),
        )

    def main(self):
        self._init_signal_listeners()
        self._init_service()

        self._loop = GLib.MainLoop()
        self._loop.run()

    def stop(self):
        self._should_stop.set()
        if self._loop:
            self._loop.quit()
            self._loop = None
            self.logger.info('Stopped D-Bus main loop')

    @staticmethod
    def _get_bus_names(bus: Bus) -> Set[str]:
        return {str(name) for name in bus.dbus.ListNames() if not name.startswith(':')}

    def path_names(
        self, bus: Bus, service: str, object_path='/', paths=None, service_dict=None
    ):
        if paths is None:
            paths = {}
        if service_dict is None:
            service_dict = {}

        paths[object_path] = {}
        try:
            obj = bus.get(service, object_path)
            interface = obj['org.freedesktop.DBus.Introspectable']
        except GLib.GError as e:
            self.logger.warning(
                f'Could not inspect D-Bus object {service}, path={object_path}: {e}'
            )
            return {}
        except KeyError as e:
            self.logger.warning(
                f'Could not get interfaces on the D-Bus object {service}, path={object_path}: {e}'
            )
            return {}

        xml_string = interface.Introspect()
        root = ElementTree.fromstring(xml_string)

        for child in root:
            if child.tag == 'node':
                if object_path == '/':
                    object_path = ''
                new_path = '/'.join((object_path, child.attrib['name']))
                self.path_names(
                    bus, service, new_path, paths, service_dict=service_dict
                )
            else:
                if not object_path:
                    object_path = '/'
                functions_dict = {}
                for func in list(child):
                    function_dict = {'name': func.attrib['name']}
                    for arg in list(func):
                        if arg.tag != 'arg':
                            continue

                        function_dict['args'] = function_dict.get('args', [])
                        function_dict['args'].append(arg.attrib)

                    if func.tag not in functions_dict:
                        functions_dict[func.tag] = []
                    functions_dict[func.tag].append(function_dict)

                if functions_dict:
                    paths[object_path][child.attrib['name']] = functions_dict

        if paths:
            service_dict[service] = paths

        return service_dict

    @action
    def query(
        self, service: Optional[str] = None, bus=tuple(t.value for t in BusType)
    ) -> Dict[str, dict]:
        """
        Query DBus for a specific service or for the full list of services.

        :param service: Service name (default: None, query all services).
        :param bus: Which bus(ses) should be queried (default: both ``system`` and ``session``).
        :return: A ``{service_name -> {properties}}`` mapping. Example:

            .. code-block:: json

              "session": {
                "org.platypush.Bus": {
                  "/": {
                    "org.freedesktop.DBus.Properties": {
                      "method": [
                        {
                          "name": "Get",
                          "args": [
                            {
                              "type": "s",
                              "name": "interface_name",
                              "direction": "in"
                            },
                            {
                              "type": "s",
                              "name": "property_name",
                              "direction": "in"
                            },
                            {
                              "type": "v",
                              "name": "value",
                              "direction": "out"
                            }
                          ]
                        },
                        {
                          "name": "GetAll",
                          "args": [
                            {
                              "type": "s",
                              "name": "interface_name",
                              "direction": "in"
                            },
                            {
                              "type": "a{sv}",
                              "name": "properties",
                              "direction": "out"
                            }
                          ]
                        },
                        {
                          "name": "Set",
                          "args": [
                            {
                              "type": "s",
                              "name": "interface_name",
                              "direction": "in"
                            },
                            {
                              "type": "s",
                              "name": "property_name",
                              "direction": "in"
                            },
                            {
                              "type": "v",
                              "name": "value",
                              "direction": "in"
                            }
                          ]
                        }
                      ],
                      "signal": [
                        {
                          "name": "PropertiesChanged",
                          "args": [
                            {
                              "type": "s",
                              "name": "interface_name"
                            },
                            {
                              "type": "a{sv}",
                              "name": "changed_properties"
                            },
                            {
                              "type": "as",
                              "name": "invalidated_properties"
                            }
                          ]
                        }
                      ]
                    },
                    "org.freedesktop.DBus.Introspectable": {
                      "method": [
                        {
                          "name": "Introspect",
                          "args": [
                            {
                              "type": "s",
                              "name": "xml_data",
                              "direction": "out"
                            }
                          ]
                        }
                      ]
                    },
                    "org.freedesktop.DBus.Peer": {
                      "method": [
                        {
                          "name": "Ping"
                        },
                        {
                          "name": "GetMachineId",
                          "args": [
                            {
                              "type": "s",
                              "name": "machine_uuid",
                              "direction": "out"
                            }
                          ]
                        }
                      ]
                    },
                    "org.platypush.Bus": {
                      "method": [
                        {
                          "name": "Post",
                          "args": [
                            {
                              "type": "s",
                              "name": "msg",
                              "direction": "in"
                            },
                            {
                              "type": "s",
                              "name": "response",
                              "direction": "out"
                            }
                          ]
                        }
                      ]
                    }
                  }
                }
              }

        """
        busses = {}
        response = {}

        if isinstance(bus, str):
            bus = (bus,)

        if BusType.SYSTEM.value in bus:
            busses['system'] = self._system_bus
        if BusType.SESSION.value in bus:
            busses['session'] = self._session_bus

        for bus_name, bus in busses.items():
            services = {}
            service_names = self._get_bus_names(bus)

            if not service:
                for srv in service_names:
                    services.update(self.path_names(bus, srv))
            elif service in service_names:
                services.update(self.path_names(bus, service))

            response[bus_name] = services

        return response

    @action
    def execute(
        self,
        service: str,
        interface: str,
        method_name: str,
        bus: str = BusType.SESSION.value,
        path: str = '/',
        args: Optional[list] = None,
    ):
        """
        Execute a method exposed on DBus.

        :param service: D-Bus service name.
        :param interface: D-Bus nterface name.
        :param method_name: Method name.
        :param bus: Bus type. Supported: ``system`` and ``session`` (default: ``session``).
        :param path: Object path.
        :param args: Arguments to be passed to the method, depending on the method signature.
        :return: Return value of the executed method.
        """
        if not args:
            args = []

        bus = self._get_bus(bus)
        obj = bus.get(service, path)[interface]
        method = getattr(obj, method_name, None)
        assert method, (
            f'No such method exposed by service={service}, '
            f'interface={interface}: {method_name}'
        )

        # Normalize any lists/dictionaries to JSON strings
        for i, arg in enumerate(args):
            if isinstance(arg, (list, tuple, dict)):
                args[i] = json.dumps(arg)

        ret = method(*args)

        try:
            ret = json.loads(json.dumps(ret))
        except Exception as e:
            self.logger.debug(e)

        return ret


# vim:sw=4:ts=4:et:



"""
Platydock is a helper script that allows you to automatically create a
Dockerfile for Platypush starting from a configuration file.
"""

import argparse
from contextlib import contextmanager
import inspect
import logging
import os
import pathlib
import re
import subprocess
import sys
import textwrap
from typing import IO, Generator, Iterable

from platypush.builder import BaseBuilder
from platypush.config import Config
from platypush.utils.manifest import (
    BaseImage,
    Dependencies,
    InstallContext,
    PackageManagers,
)

logger = logging.getLogger()


class DockerBuilder(BaseBuilder):
    """
    Creates a Platypush Docker image from a configuration file.
    """

    _pkg_manager_by_base_image = {
        BaseImage.ALPINE: PackageManagers.APK,
        BaseImage.DEBIAN: PackageManagers.APT,
        BaseImage.FEDORA: PackageManagers.DNF,
        BaseImage.UBUNTU: PackageManagers.APT,
    }

    _header = textwrap.dedent(
        """
        # This Dockerfile was automatically generated by Platydock.
        #
        # You can build a Platypush image from it by running
        # `docker build -t platypush .` in the same folder as this file,
        # or copy it to the root a Platypush source folder to install the
        # checked out version instead of downloading it first.
        #
        # You can then run your new image through:
        #   docker run --rm --name platypush \\
        #       -v /path/to/your/config/dir:/etc/platypush \\
        #       -v /path/to/your/workdir:/var/lib/platypush \\
        #       -p 8008:8008 \\
        #       platypush\n
        """
    )

    _footer = textwrap.dedent(
        """
        # You can customize the name of your installation by passing
        # --device-id=... to the launched command.
        """
    )

    def __init__(
        self, *args, image: BaseImage, tag: str, print_only: bool = False, **kwargs
    ):
        kwargs['install_context'] = InstallContext.DOCKER
        super().__init__(*args, **kwargs)
        self.image = image
        self.tag = tag
        self.print_only = print_only

    @classmethod
    def get_name(cls):
        return "platydock"

    @classmethod
    def get_description(cls):
        return "Build a Platypush Docker image from a configuration file."

    @property
    def dockerfile_dir(self) -> str:
        """
        Proxy property for the output Dockerfile directory.
        """
        output = self.output
        parent = os.path.dirname(output)

        if os.path.isfile(output):
            return parent

        if os.path.isdir(output):
            return output

        logger.info('%s directory does not exist, creating it', output)
        pathlib.Path(output).mkdir(mode=0o750, parents=True, exist_ok=True)
        return output

    @property
    def dockerfile(self) -> str:
        """
        Proxy property for the output Dockerfile.
        """
        return os.path.join(self.dockerfile_dir, 'Dockerfile')

    @property
    def pkg_manager(self) -> PackageManagers:
        """
        Proxy property for the package manager to use.
        """
        return self._pkg_manager_by_base_image[self.image]

    def _read_base_dockerfile_lines(self) -> Generator[str, None, None]:
        """
        :return: The lines of the base Dockerfile.
        """
        import platypush

        base_file = os.path.join(
            str(pathlib.Path(inspect.getfile(platypush)).parent),
            'install',
            'docker',
            f'{self.image}.Dockerfile',
        )

        with open(base_file, 'r') as f:
            for line in f:
                yield line.rstrip()

    @property
    def deps(self) -> Dependencies:
        return Dependencies.from_config(
            self.cfgfile,
            pkg_manager=self.pkg_manager,
            install_context=InstallContext.DOCKER,
            base_image=self.image,
        )

    def _create_dockerfile_parser(self):
        """
        Closure for a context-aware parser for the default Dockerfile.
        """
        is_after_expose_cmd = False
        deps = self.deps
        ports = self._get_exposed_ports()

        def parser():
            nonlocal is_after_expose_cmd

            for line in self._read_base_dockerfile_lines():
                if re.search(r'\s*rm -rf /install\s*', line):
                    for new_line in [*deps.to_pip_install_commands(), *deps.after]:
                        yield f'\t{new_line} && \\'
                elif line.startswith('EXPOSE ') and ports:
                    if not is_after_expose_cmd:
                        yield from [f'EXPOSE {port}' for port in ports]
                        is_after_expose_cmd = True

                    continue
                elif line.startswith('CMD'):
                    yield from self._footer.split('\n')

                yield line

                if line.startswith('RUN /install/platypush/install/scripts/'):
                    for new_line in [*deps.before, *deps.to_pkg_install_commands()]:
                        yield f'\t{new_line} && \\'

                if line.startswith('CMD') and self.device_id:
                    yield f'\t--device-id {self.device_id} \\'

        return parser

    def build(self):
        """
        Build a Dockerfile based on a configuration file.

        :return: The content of the generated Dockerfile.
        """

        # Set the DOCKER_CTX environment variable so any downstream logic knows
        # that we're running in a Docker build context.
        os.environ['DOCKER_CTX'] = '1'

        self._generate_dockerfile()
        if self.print_only:
            return

        self._build_image()
        self._print_instructions(
            textwrap.dedent(
                f"""
                A Docker image has been built from the configuration file {self.cfgfile}.
                The Dockerfile is available under {self.dockerfile}.
                You can run the Docker image with the following command:

                    docker run \\
                        --rm --name platypush \\
                        -v {os.path.dirname(self.cfgfile)}:/etc/platypush \\
                        -v /path/to/your/workdir:/var/lib/platypush \\
                        -p 8008:8008 \\
                        platypush
                """
            )
        )

    def _build_image(self):
        """
        Build a Platypush Docker image from the generated Dockerfile.
        """
        logger.info('Building Docker image...')
        cmd = [
            'docker',
            'build',
            '-f',
            self.dockerfile,
            '-t',
            self.tag,
            '.',
        ]

        subprocess.run(cmd, check=True)

    def _generate_dockerfile(self):
        """
        Parses the configuration file and generates a Dockerfile based on it.
        """

        @contextmanager
        def open_writer() -> Generator[IO, None, None]:
            # flake8: noqa
            f = sys.stdout if self.print_only else open(self.dockerfile, 'w')

            try:
                yield f
            finally:
                if f is not sys.stdout:
                    f.close()

        if not self.print_only:
            logger.info('Parsing configuration file %s...', self.cfgfile)

        Config.init(self.cfgfile)

        if not self.print_only:
            logger.info('Generating Dockerfile %s...', self.dockerfile)

        parser = self._create_dockerfile_parser()

        with open_writer() as f:
            f.write(self._header + '\n')
            for line in parser():
                f.write(line + '\n')

    @classmethod
    def _get_arg_parser(cls) -> argparse.ArgumentParser:
        parser = super()._get_arg_parser()

        parser.add_argument(
            '-i',
            '--image',
            dest='image',
            required=False,
            type=BaseImage,
            choices=list(BaseImage),
            default=BaseImage.ALPINE,
            help='Base image to use for the Dockerfile (default: alpine).',
        )

        parser.add_argument(
            '-t',
            '--tag',
            dest='tag',
            required=False,
            type=str,
            default='platypush:latest',
            help='Tag name to be used for the built image '
            '(default: "platypush:latest").',
        )

        parser.add_argument(
            '--print',
            dest='print_only',
            action='store_true',
            help='Use this flag if you only want to print the Dockerfile to '
            'stdout instead of generating an image.',
        )

        return parser

    @staticmethod
    def _get_exposed_ports() -> Iterable[int]:
        """
        :return: The listen ports used by the backends enabled in the configuration
            file.
        """
        backends_config = Config.get_backends()
        return {
            int(port)
            for port in (
                backends_config.get('http', {}).get('port'),
                backends_config.get('tcp', {}).get('port'),
            )
            if port
        }


def main():
    """
    Generates a Dockerfile based on the configuration file.
    """
    DockerBuilder.from_cmdline(sys.argv[1:]).build()
    return 0


if __name__ == '__main__':
    sys.exit(main())


# vim:sw=4:ts=4:et:



from marshmallow import fields
from marshmallow.schema import Schema


class PingResponseSchema(Schema):
    """
    Ping response schema.
    """

    host = fields.String(
        required=True,
        metadata={
            "description": "Remote host IP or name",
            "example": "platypush.tech",
        },
    )

    success = fields.Boolean(
        required=True,
        metadata={
            "description": "True if the ping was successful, False otherwise",
            "example": True,
        },
    )

    min = fields.Float(
        required=False,
        metadata={
            "description": "Minimum round-trip time (in ms)",
            "example": 0.1,
        },
    )

    max = fields.Float(
        required=False,
        metadata={
            "description": "Maximum round-trip time (in ms)",
            "example": 0.2,
        },
    )

    avg = fields.Float(
        required=False,
        metadata={
            "description": "Average round-trip time (in ms)",
            "example": 0.15,
        },
    )

    mdev = fields.Float(
        required=False,
        metadata={
            "description": "Standard deviation of the round-trip time (in ms)",
            "example": 0.05,
        },
    )



import asyncio
import logging
import threading

from abc import ABC, abstractmethod
from functools import wraps
from typing import Any, Callable, Optional

from platypush.bus import Bus
from platypush.common import ExtensionWithManifest
from platypush.event import EventGenerator
from platypush.message.response import Response
from platypush.utils import get_decorators, get_plugin_name_by_class

PLUGIN_STOP_TIMEOUT = 5  # Plugin stop timeout in seconds
_logger = logging.getLogger(__name__)


def action(f: Callable[..., Any]) -> Callable[..., Response]:
    """
    Decorator used to wrap the methods in the plugin classes that should be
    exposed as actions.

    It wraps the method's response into a generic
    :meth:`platypush.message.response.Response` object.
    """

    @wraps(f)
    def _execute_action(*args, **kwargs) -> Response:
        response = Response()
        try:
            result = f(*args, **kwargs)
        except Exception as e:
            if isinstance(e, KeyboardInterrupt):
                return response

            _logger.exception(e)
            result = Response(errors=[str(e)])

        if result and isinstance(result, Response):
            result.errors = (
                result.errors if isinstance(result.errors, list) else [result.errors]
            )
            response = result
        elif isinstance(result, tuple) and len(result) == 2:
            response.errors = result[1] if isinstance(result[1], list) else [result[1]]

            if len(response.errors) == 1 and response.errors[0] is None:
                response.errors = []
            response.output = result[0]
        else:
            response = Response(output=result, errors=[])

        return response

    # Propagate the docstring
    _execute_action.__doc__ = f.__doc__
    # Expose the wrapped function
    _execute_action.wrapped = f  # type: ignore
    return _execute_action


class Plugin(EventGenerator, ExtensionWithManifest):  # lgtm [py/missing-call-to-init]
    """Base plugin class"""

    def __init__(self, **kwargs):
        super().__init__()

        self.logger = logging.getLogger(
            'platypush:plugin:' + get_plugin_name_by_class(self.__class__)
        )
        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs['logging'].upper()))

        self.registered_actions = set(
            get_decorators(self.__class__, climb_class_hierarchy=True).get('action', [])
        )

    @property
    def _db(self):
        """
        :return: The reference to the :class:`platypush.plugins.db.DbPlugin`.
        """
        from platypush.context import get_plugin
        from platypush.plugins.db import DbPlugin

        db: DbPlugin = get_plugin(DbPlugin)  # type: ignore
        assert db, 'db plugin not initialized'
        return db

    @property
    def _redis(self):
        """
        :return: The reference to the :class:`platypush.plugins.redis.RedisPlugin`.
        """
        from platypush.context import get_plugin
        from platypush.plugins.redis import RedisPlugin

        redis: RedisPlugin = get_plugin(RedisPlugin)  # type: ignore
        assert redis, 'db plugin not initialized'
        return redis

    @property
    def _bus(self):
        """
        :return: The reference to the :class:`platypush.bus.Bus`.
        """
        from platypush.context import get_bus

        return get_bus()

    @property
    def _entities(self):
        """
        :return: The reference to the :class:`platypush.plugins.entities.EntitiesPlugin`.
        """
        from platypush.context import get_plugin
        from platypush.plugins.entities import EntitiesPlugin

        entities: EntitiesPlugin = get_plugin('entities')  # type: ignore
        assert entities, 'entities plugin not initialized'
        return entities

    def __str__(self):
        """
        :return: The qualified name of the plugin.
        """
        return get_plugin_name_by_class(self.__class__)

    def run(self, method, *args, **kwargs):
        assert (
            method in self.registered_actions
        ), f'{method} is not a registered action on {self.__class__.__name__}'
        return getattr(self, method)(*args, **kwargs)


class RunnablePlugin(Plugin):
    """
    Class for runnable plugins - i.e. plugins that have a start/stop method and can be started.
    """

    def __init__(
        self,
        poll_interval: Optional[float] = 15,
        stop_timeout: Optional[float] = PLUGIN_STOP_TIMEOUT,
        disable_monitor: bool = False,
        **kwargs,
    ):
        """
        :param poll_interval: How often the :meth:`.loop` function should be
            executed (default: 15 seconds). *NOTE*: For back-compatibility
            reasons, the `poll_seconds` argument is also supported, but it's
            deprecated.
        :param stop_timeout: How long we should wait for any running
            threads/processes to stop before exiting (default: 5 seconds).
        :param disable_monitor: If set to True then the plugin will not monitor
            for new events. This is useful if you want to run a plugin in
            stateless mode and only leverage its actions, without triggering any
            events. Defaults to False.
        """
        super().__init__(**kwargs)
        self.poll_interval = poll_interval
        self.bus: Optional[Bus] = None
        self.disable_monitor = disable_monitor
        self._should_stop = threading.Event()
        self._stop_timeout = stop_timeout
        self._thread: Optional[threading.Thread] = None

        if kwargs.get('poll_seconds') is not None:
            self.logger.warning('poll_seconds is deprecated, use poll_interval instead')

            if self.poll_interval is None:
                self.poll_interval = kwargs['poll_seconds']

    def main(self):
        """
        Implementation of the main loop of the plugin.
        """
        raise NotImplementedError()

    def should_stop(self) -> bool:
        return self._should_stop.is_set()

    def wait_stop(self, timeout=None):
        """
        Wait until a stop event is received.
        """
        if self.disable_monitor:
            # Wait indefinitely if the monitor is disabled
            return self._should_stop.wait(timeout=None)

        return self._should_stop.wait(timeout=timeout)

    def start(self):
        """
        Start the plugin.
        """
        self._thread = threading.Thread(
            target=self._runner, name=self.__class__.__name__
        )
        self._thread.start()

    def stop(self):
        """
        Stop the plugin.
        """
        self._should_stop.set()
        if (
            self._thread
            and self._thread != threading.current_thread()
            and self._thread.is_alive()
        ):
            self.logger.info('Waiting for the plugin to stop')
            try:
                if self._thread:
                    self._thread.join(timeout=self._stop_timeout)
                    if self._thread and self._thread.is_alive():
                        self.logger.warning(
                            'Timeout (seconds=%s) on exit for the plugin',
                            self._stop_timeout,
                        )
            except Exception as e:
                self.logger.warning('Could not join thread on stop: %s', e)

        self.logger.info(
            'Stopped plugin: [%s]', get_plugin_name_by_class(self.__class__)
        )

    def _runner(self):
        """
        Implementation of the runner thread.
        """
        if self.disable_monitor:
            return

        self.logger.info(
            'Starting plugin: [%s]', get_plugin_name_by_class(self.__class__)
        )

        while not self.should_stop():
            try:
                self.main()
            except Exception as e:
                self.logger.exception(e)

            if self.poll_interval:
                self.wait_stop(self.poll_interval)

        self._thread = None


class AsyncRunnablePlugin(RunnablePlugin, ABC):
    """
    Class for runnable plugins with an asynchronous event loop attached.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._loop: Optional[asyncio.AbstractEventLoop] = asyncio.new_event_loop()
        self._task: Optional[asyncio.Task] = None

    @property
    def _should_start_runner(self):
        """
        This property is used to determine if the runner and the event loop
        should be started for this plugin.
        """
        return True

    @abstractmethod
    async def listen(self):
        """
        Main body of the async plugin. When it's called, the event loop should
        already be running and available over `self._loop`.
        """

    async def _listen(self):
        """
        Wrapper for :meth:`.listen` that catches any exceptions and logs them.
        """
        try:
            await self.listen()
        except KeyboardInterrupt:
            pass
        except RuntimeError as e:
            if not (
                str(e).startswith('Event loop stopped before ')
                or str(e).startswith('no running event loop')
            ):
                raise e

    def _run_listener(self):
        """
        Initialize an event loop and run the listener as a task.
        """
        assert self._loop, 'The loop is not initialized'
        asyncio.set_event_loop(self._loop)

        self._task = self._loop.create_task(self._listen())
        if hasattr(self._task, 'set_name'):
            self._task.set_name(self.__class__.__name__ + '.listen')
        try:
            self._loop.run_until_complete(self._task)
        except Exception as e:
            if not self.should_stop():
                self.logger.warning('The loop has terminated with an error')
                self.logger.exception(e)

        self._task.cancel()

    def main(self):
        if self.should_stop():
            self.logger.info('The plugin is already scheduled to stop')
            return

        self._loop = asyncio.new_event_loop()

        if self._should_start_runner:
            while not self.should_stop():
                try:
                    self._run_listener()
                finally:
                    self.wait_stop(self.poll_interval)
        else:
            self.wait_stop()

    def stop(self):
        if self._loop and self._loop.is_running():
            self._loop.call_soon_threadsafe(self._loop.stop)
            self._loop = None

        super().stop()


# vim:sw=4:ts=4:et:



import base64
import datetime
import errno
import getpass
import os
import threading

from binascii import hexlify
from stat import (
    S_ISDIR,
    S_ISREG,
    S_ISLNK,
    S_ISCHR,
    S_ISFIFO,
    S_ISSOCK,
    S_ISBLK,
    S_ISDOOR,
)
from typing import Optional, Dict, Tuple, List, Union, Any

from paramiko import DSSKey, RSAKey, SSHClient, WarningPolicy, SFTPClient

try:
    from paramiko.util import u
except ImportError:
    from paramiko.py3compat import u  # type: ignore

try:
    from paramiko import GSS_AUTH_AVAILABLE  # type: ignore
except ImportError:
    from paramiko.ssh_gss import GSS_AUTH_AVAILABLE

from platypush import Response
from platypush.plugins import Plugin, action
from platypush.plugins.ssh.tunnel.forward import forward_tunnel
from platypush.plugins.ssh.tunnel.reverse import reverse_tunnel, close_tunnel


class SshPlugin(Plugin):
    """
    SSH plugin.
    """

    key_dispatch_table = {'dsa': DSSKey, 'rsa': RSAKey}

    def __init__(
        self, key_file: Optional[str] = None, passphrase: Optional[str] = None, **kwargs
    ):
        """
        :param key_file: Default key file (default: any "id_rsa", "id_dsa",
            "id_ecdsa", or "id_ed25519" key discoverable in ``~/.ssh/``.
        :param passphrase: Key file passphrase (default: None).
        """
        super().__init__(**kwargs)
        self.key_file = (
            os.path.abspath(os.path.expanduser(key_file)) if key_file else None
        )
        self.passphrase = passphrase
        self._sessions: Dict[Tuple[str, int, Optional[str]], SSHClient] = {}
        self._fwd_tunnels: Dict[Tuple[int, str, int], dict] = {}
        self._rev_tunnels: Dict[Tuple[int, str, int], dict] = {}

    def _get_key(
        self, key_file: Optional[str] = None, passphrase: Optional[str] = None
    ):
        key_file = key_file or self.key_file
        return (
            os.path.abspath(os.path.expanduser(key_file)) if key_file else None,
            passphrase or self.passphrase,
        )

    @staticmethod
    def _get_host_port_user(host: str, port: int = 22, user: Optional[str] = None, **_):
        if host.find('@') >= 0:
            user, host = host.split('@')
        if host.find(':') >= 0:
            host, p = host.split(':')
            port = int(p)
        if not user:
            user = getpass.getuser()

        return host, port, user

    @action
    def keygen(
        self,
        filename: str,
        type: str = 'rsa',  # pylint: disable=redefined-builtin
        bits: int = 4096,
        comment: Optional[str] = None,
        passphrase: Optional[str] = None,
    ) -> dict:
        """
        Generate an SSH keypair.

        :param filename: Output file name for the private key (the public key
            will be stored in <filename>.pub).
        :param type: Encryption algorithm, either "rsa" or "dsa" (default: "rsa").
        :param bits: Key length in bits (default: 4096).
        :param comment: Key comment (default: None).
        :param passphrase: Key passphrase (default: None).
        :return:

            .. code-block:: json

              {
                "fingerprint": "SHA256:...",
                "key_file": "private_key_file",
                "pub_key_file": "public_key_file"
              }

        """
        assert type != 'dsa' or bits <= 1024, 'DSA keys support a maximum of 1024 bits'
        assert (
            type in self.key_dispatch_table
        ), f'No such type: {type}. Available types: {self.key_dispatch_table.keys()}'

        if filename:
            filename = os.path.abspath(os.path.expanduser(filename))

        prv = self.key_dispatch_table[type].generate(bits=bits)
        prv.write_private_key_file(filename=filename, password=passphrase)
        pub = self.key_dispatch_table[type](filename=filename, password=passphrase)
        pub_file = f'{filename}.pub'
        with open(pub_file, 'w') as f:
            f.write(f'{pub.get_name()} {pub.get_base64()}')
            if comment:
                f.write(' ' + comment)

        return {
            'fingerprint': u(hexlify(pub.get_fingerprint())),
            'key_file': filename,
            'pub_key_file': pub_file,
        }

    def run(self, *args, **kwargs):
        try:
            return super().run(*args, **kwargs)
        except Exception as e:
            raise AssertionError(e) from e

    def _connect(
        self,
        host: str,
        port: int = 22,
        user: Optional[str] = None,
        password: Optional[str] = None,
        key_file: Optional[str] = None,
        passphrase: Optional[str] = None,
        compress: bool = False,
        timeout: Optional[int] = None,
        auth_timeout: Optional[int] = None,
    ) -> SSHClient:
        try:
            host, port, user = self._get_host_port_user(host, port, user)
            key = (host, port, user)
            if key in self._sessions:
                self.logger.info(
                    '[Connect] The SSH session is already active: %s@%s:%d',
                    user,
                    host,
                    port,
                )
                return self._sessions[key]

            key_file, passphrase = self._get_key(key_file, passphrase)
            client = SSHClient()
            client.set_missing_host_key_policy(WarningPolicy())

            args = {
                'hostname': host,
                'port': port,
                'username': user,
                'compress': compress,
                'timeout': timeout,
                'auth_timeout': auth_timeout,
            }

            if password:
                args['password'] = password
            elif key_file:
                args['key_filename'] = key_file
                args['passphrase'] = passphrase
                args['gss_auth'] = GSS_AUTH_AVAILABLE
                args['gss_kex'] = GSS_AUTH_AVAILABLE
            else:
                client.load_system_host_keys()

            client.connect(**args)
            self._sessions[key] = client
            return client
        except Exception as e:
            self.logger.exception(e)
            raise AssertionError(f'Connection to {host} failed: {type(e)}: {e}')

    @action
    def connect(
        self,
        host: str,
        port: int = 22,
        user: Optional[str] = None,
        password: Optional[str] = None,
        key_file: Optional[str] = None,
        passphrase: Optional[str] = None,
        compress: bool = False,
        timeout: Optional[int] = None,
        auth_timeout: Optional[int] = None,
    ) -> None:
        """
        Open an SSH connection.

        :param host: Host name or IP. Can also be in the format ``[user]@<host>:[port]``.
        :param port: Remote port (default: 22).
        :param user: Username (default: None, same user name as the one running platypush).
        :param password: Password (default: None).
        :param key_file: Key file to use for authentication (default: None).
        :param passphrase: Passphrase for the key file (default: None).
        :param compress: Compress data on the connection (default: False).
        :param timeout: Data transfer timeout in seconds (default: None).
        :param auth_timeout: Authentication timeout in seconds (default: None).
        """
        self._connect(
            host=host,
            port=port,
            user=user,
            password=password,
            key_file=key_file,
            passphrase=passphrase,
            compress=compress,
            timeout=timeout,
            auth_timeout=auth_timeout,
        )

    @action
    def disconnect(self, host: str, port: int = 22, user: Optional[str] = None) -> None:
        """
        Close a connection to a host.

        :param host: Host name or IP. Can also be in the format ``[user]@<host>:[port]``.
        :param port: Remote port (default: 22).
        :param user: Username (default: None, same user name as the one running platypush).
        """
        host, port, user = self._get_host_port_user(host, port, user)
        key = (host, port, user)
        if key not in self._sessions:
            self.logger.info(
                '[Disconnect] The SSH session is not active: %s@%s:%d', user, host, port
            )

        session = self._sessions[key]
        try:
            session.close()
        except Exception as e:
            self.logger.exception(e)

        del self._sessions[key]

    @action
    def exec(
        self,
        cmd: str,
        keep_alive: bool = False,
        timeout: Optional[int] = None,
        stdin: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Response:
        """
        Run a command on a host.

        :param cmd: Command to run
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param timeout: Communication timeout in seconds (default: None).
        :param stdin: Optional string to pass on the stdin of the command.
        :param env: Dictionary of environment variables to be used for the connection (default: None).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        :return: The output of the executed command.
        """
        client = self._connect(**kwargs)

        def decode(buf: bytes) -> str:
            try:
                s_buf = buf.decode()
            except (ValueError, TypeError):
                s_buf = base64.encodebytes(buf).decode()

            if s_buf.endswith('\n'):
                s_buf = s_buf[:-1]
            return s_buf

        try:
            _in, _out, _err = client.exec_command(cmd, timeout=timeout, environment=env)
            if stdin:
                with _in:
                    _in.write(stdin)

            resp = Response()
            with _out:
                resp.output = decode(_out.read())
            with _err:
                err = decode(_err.read())
                if err:
                    resp.errors = [err]

            return resp
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @staticmethod
    def is_directory(sftp: SFTPClient, path: str) -> bool:
        f = sftp.lstat(path)
        if f.st_mode is None:
            return False

        return S_ISDIR(f.st_mode)

    @classmethod
    def sftp_walk(cls, sftp: SFTPClient, path: str):
        files = []
        folders = []

        for f in sftp.listdir_attr(path):
            if f.st_mode is not None and S_ISDIR(f.st_mode):
                folders.append(f.filename)
            else:
                files.append(f.filename)

        yield path, folders, files

        for folder in folders:
            new_path = os.path.join(path, folder)
            yield from cls.sftp_walk(sftp, new_path)

    def sftp_get(
        self,
        sftp: SFTPClient,
        remote_path: str,
        local_path: str,
        recursive: bool = False,
    ) -> None:
        if self.is_directory(sftp, remote_path):
            assert (
                recursive
            ), '{} is a directory on the server but recursive has been set to False'
            local_path = os.path.join(local_path, os.path.basename(remote_path))
            os.makedirs(local_path, mode=0o755, exist_ok=True)
            sftp.chdir(remote_path)

            for path, _, files in self.sftp_walk(sftp, '.'):
                new_local_path = os.path.join(local_path, path)
                os.makedirs(new_local_path, mode=0o755, exist_ok=True)

                for file in files:
                    self.logger.info(
                        'Downloading file %s from %s to %s', file, path, new_local_path
                    )

                    self.sftp_get(
                        sftp,
                        os.path.join(remote_path, path, file),
                        os.path.join(new_local_path, file),
                        recursive=recursive,
                    )
        else:
            if os.path.isdir(local_path):
                local_path = os.path.join(local_path, os.path.basename(remote_path))

            sftp.get(remote_path, local_path)

    @action
    def get(
        self,
        remote_path: str,
        local_path: str,
        recursive: bool = False,
        keep_alive: bool = False,
        **kwargs,
    ) -> None:
        """
        Download a file or folder from an SSH server.

        :param remote_path: Remote path (file or directory).
        :param local_path: Local path (file or directory).
        :param recursive: Set to True if you want to recursively download folders (default: False).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        local_path = os.path.abspath(os.path.expanduser(local_path))
        kwargs['compress'] = True
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            self.sftp_get(
                sftp,
                remote_path=remote_path,
                local_path=local_path,
                recursive=recursive,
            )
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def put(
        self,
        remote_path: str,
        local_path: str,
        recursive: bool = False,
        keep_alive: bool = False,
        **kwargs,
    ) -> None:
        """
        Upload a file or folder to an SSH server.

        :param remote_path: Remote path (file or directory).
        :param local_path: Local path (file or directory).
        :param recursive: Set to True if you want to recursively upload folders (default: False).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        local_path = os.path.abspath(os.path.expanduser(local_path))
        assert os.path.exists(local_path), os.strerror(errno.ENOENT)
        kwargs['compress'] = True
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            if os.path.isdir(local_path):
                try:
                    sftp.mkdir(remote_path)
                except Exception as e:
                    self.logger.warning(
                        'mkdir %s failed: %s: %s', remote_path, type(e), e
                    )

                assert (
                    recursive
                ), f'{local_path} is a directory but recursive has been set to False'

                assert self.is_directory(
                    sftp, remote_path
                ), f'{remote_path} is not a directory on the remote host'

                sftp.chdir(remote_path)
                os.chdir(local_path)

                for path, _, files in os.walk('.'):
                    try:
                        sftp.mkdir(path)
                    except Exception as e:
                        self.logger.warning(
                            'mkdir %s failed: %s: %s', remote_path, type(e), e
                        )

                    for file in files:
                        src = os.path.join(path, file)
                        dst = os.path.join(path, file)
                        self.logger.info('Copying %s to %s', src, dst)
                        sftp.put(src, dst)
            else:
                if self.is_directory(sftp, remote_path):
                    remote_path = os.path.join(
                        remote_path, os.path.basename(local_path)
                    )

                sftp.put(local_path, remote_path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def ls(
        self, path: str = '.', attrs: bool = False, keep_alive: bool = False, **kwargs
    ) -> Union[List[str], Dict[str, Any]]:
        """
        Return the list of files in a path on a remote server.

        :param path: Remote path (default: current directory).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param attrs: Set to True if you want to get the full information of each file (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        :return: A list of filenames if ``attrs=False``, otherwise a dictionary ``filename -> {attributes`` if
            ``attrs=True``.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        def get_file_type(st_mode: Optional[int]) -> str:
            if st_mode is None:
                return 'unknown'
            if S_ISDIR(st_mode):
                return 'directory'
            elif S_ISBLK(st_mode):
                return 'block'
            elif S_ISCHR(st_mode):
                return 'device'
            elif S_ISDOOR(st_mode):
                return 'door'
            elif S_ISREG(st_mode):
                return 'file'
            elif S_ISLNK(st_mode):
                return 'link'
            elif S_ISFIFO(st_mode):
                return 'fifo'
            elif S_ISSOCK(st_mode):
                return 'sock'
            else:
                return 'unknown'

        try:
            if attrs:
                return {
                    f.filename: {
                        'filename': f.filename,
                        'longname': f.longname,
                        'attributes': f.attr,
                        'type': get_file_type(f.st_mode),
                        'access_time': (
                            datetime.datetime.fromtimestamp(f.st_atime)
                            if f.st_atime
                            else None
                        ),
                        'modify_time': (
                            datetime.datetime.fromtimestamp(f.st_mtime)
                            if f.st_mtime
                            else None
                        ),
                        'uid': f.st_uid,
                        'gid': f.st_gid,
                        'size': f.st_size,
                    }
                    for f in sftp.listdir_attr(path)
                }

            return sftp.listdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def rm(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Remove a file from the server.

        :param path: Remote path to remove.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.remove(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def mv(self, path: str, new_path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Move/rename a file.

        :param path: Remote path to move/rename.
        :param new_path: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.posix_rename(path, new_path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def mkdir(
        self, path: str, mode: int = 0o777, keep_alive: bool = False, **kwargs
    ) -> None:
        """
        Create a directory.

        :param path: Path to be created.
        :param mode: Access permissions (default: 0777).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.mkdir(path, mode=mode)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def rmdir(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Remove a directory.

        :param path: Path to be removed.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.rmdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def ln(self, src: str, dest: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Create a symbolic link.

        :param src: Source path.
        :param dest: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.symlink(src, dest)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chmod(self, path: str, mode: int, keep_alive: bool = False, **kwargs) -> None:
        """
        Change the access rights of a path.

        :param path: Path to be modified.
        :param mode: Access permissions (in octal mode).
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chmod(path, mode=mode)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chown(
        self, path: str, uid: int, gid: int, keep_alive: bool = False, **kwargs
    ) -> None:
        """
        Change the owner of a path.

        :param path: Path to be modified.
        :param uid: New user ID.
        :param gid: New group ID.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chown(path, uid=uid, gid=gid)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def chdir(self, path: str, keep_alive: bool = False, **kwargs) -> None:
        """
        Change directory to the specified path.

        :param path: Destination path.
        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            sftp.chdir(path)
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def getcwd(self, keep_alive: bool = False, **kwargs) -> str:
        """
        Get the current working directory.

        :param keep_alive: Keep the connection active after running the command (default: False).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        client = self._connect(**kwargs)
        sftp = client.open_sftp()

        try:
            return sftp.getcwd() or '/'
        finally:
            if not keep_alive:
                host, port, user = self._get_host_port_user(**kwargs)
                self.disconnect(host=host, port=port, user=user)

    @action
    def start_forward_tunnel(
        self,
        local_port: int,
        remote_host: str,
        remote_port: int,
        bind_addr: str = '',
        **kwargs,
    ):
        """
        Start an SSH forward tunnel, tunnelling ``<local_port>`` to
        ``<remote_host>:<remote_port>``.

        :param local_port: Local port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        :param bind_addr: If set, the `local_port` will be bound to this
            address/subnet (default: '', or 0.0.0.0: any).
        :param kwargs: Arguments for
            :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        key = local_port, remote_host, remote_port
        if key in self._fwd_tunnels:
            self.logger.info(
                'The tunnel %s:%d:%s:%d is already active',
                bind_addr,
                local_port,
                remote_host,
                remote_port,
            )
            return

        client = self._connect(**kwargs)
        server = forward_tunnel(
            local_port,
            remote_host,
            remote_port,
            client.get_transport(),
            bind_addr=bind_addr,
        )
        threading.Thread(target=server.serve_forever, name='sshfwdtun').start()

        self._fwd_tunnels[key] = {
            'client': client,
            'server': server,
            'args': kwargs,
        }

    @action
    def stop_forward_tunnel(self, local_port: int, remote_host: str, remote_port: int):
        """
        Stop an active SSH forward tunnel.

        :param local_port: Local port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        """
        key = (local_port, remote_host, remote_port)
        if key not in self._fwd_tunnels:
            self.logger.warning(
                'No such forward tunnel: %d:%s:%d', local_port, remote_host, remote_port
            )
            return

        server = self._fwd_tunnels[key]['server']
        server.server_close()

        args = self._fwd_tunnels[key]['args']
        host, port, user = self._get_host_port_user(**args)
        self.disconnect(host=host, port=port, user=user)

    @action
    def start_reverse_tunnel(
        self,
        server_port: int,
        remote_host: str,
        remote_port: int,
        bind_addr: str = '',
        **kwargs,
    ):
        """
        Start an SSH reversed tunnel. <server_port> on the SSH server is forwarded across an SSH session back to the
        local machine, and out to a <remote_host>:<remote_port> reachable from this network.

        :param server_port: Server port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        :param bind_addr: If set, the `server_port` will be bound to this address/subnet (default: '', or 0.0.0.0: any).
        :param kwargs: Arguments for :meth:`platypush.plugins.ssh.SshPlugin.connect`.
        """
        key = server_port, remote_host, remote_port
        if key in self._fwd_tunnels:
            self.logger.info(
                'The tunnel %s:%d:%s:%d is already active',
                bind_addr,
                server_port,
                remote_host,
                remote_port,
            )
            return

        client = self._connect(**kwargs)
        server = reverse_tunnel(
            server_port,
            remote_host,
            remote_port,
            transport=client.get_transport(),
            bind_addr=bind_addr,
        )

        threading.Thread(target=server, name='sshrevtun').start()

        self._rev_tunnels[key] = {
            'client': client,
            'server': server,
            'args': kwargs,
        }

    @action
    def stop_reverse_tunnel(self, server_port: int, remote_host: str, remote_port: int):
        """
        Stop an active SSH reversed tunnel.

        :param server_port: Server port.
        :param remote_host: Remote host.
        :param remote_port: Remote port.
        """
        key = (server_port, remote_host, remote_port)
        if key not in self._rev_tunnels:
            self.logger.warning(
                'No such reverse tunnel: %d:%s:%d',
                server_port,
                remote_host,
                remote_port,
            )
            return

        close_tunnel(*key)
        args = self._rev_tunnels[key]['args']
        host, port, user = self._get_host_port_user(**args)
        self.disconnect(host=host, port=port, user=user)


# vim:sw=4:ts=4:et:



import os
import subprocess
from typing import Optional, Tuple

from platypush.plugins import action
from platypush.plugins.camera import CameraPlugin, Camera


class CameraIrMlx90640Plugin(CameraPlugin):
    """
    Plugin to interact with a `ML90640 <https://shop.pimoroni.com/products/mlx90640-thermal-camera-breakout>`_
    infrared thermal camera.

    In order to use this plugin you'll need to download and compile the
    `mlx90640 <https://github.com/pimoroni/mlx90640-library>`_ C++ bindings and examples for the device.
    Instructions on Raspbian::

        # Install the dependencies
        $ [sudo] apt-get install libi2c-dev
        $ cd $PLATYPUSH_SRC_DIR
        $ git submodule init
        $ git submodule update
        $ cd platypush/plugins/camera/ir/mlx90640/lib
        $ make clean
        $ make bcm2835
        $ make examples/rawrgb I2C_MODE=LINUX

    """

    def __init__(
        self,
        rawrgb_path: Optional[str] = None,
        resolution: Tuple[int, int] = (32, 24),
        warmup_frames: Optional[int] = 5,
        **kwargs
    ):
        """
        :param rawrgb_path: Specify it if the rawrgb executable compiled from
            https://github.com/pimoroni/mlx90640-library is in another folder than
            `<directory of this file>/lib/examples`.
        :param resolution: Device resolution (default: 32x24).
        :param warmup_frames: Number of frames to be skipped on sensor initialization/warmup (default: 2).
        :param kwargs: Extra parameters to be passed to :class:`platypush.plugins.camera.CameraPlugin`.
        """
        super().__init__(
            device='mlx90640',
            resolution=resolution,
            warmup_frames=warmup_frames,
            **kwargs
        )

        if not rawrgb_path:
            rawrgb_path = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), 'lib', 'examples', 'rawrgb'
            )
        rawrgb_path = os.path.abspath(os.path.expanduser(rawrgb_path))

        assert os.path.isfile(
            rawrgb_path
        ), 'rawrgb executable not found. Please follow the documentation of this plugin to build it'

        self.rawrgb_path = rawrgb_path
        self._capture_proc = None

    def _is_capture_running(self):
        return self._capture_proc is not None and self._capture_proc.poll() is None

    def prepare_device(self, device: Camera, **_):
        if not self._is_capture_running():
            self._capture_proc = subprocess.Popen(
                [self.rawrgb_path, '{}'.format(device.info.fps)],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )

        return self._capture_proc

    def release_device(self, device: Camera):
        if not self._is_capture_running():
            return

        self._capture_proc.terminate()
        self._capture_proc.kill()
        self._capture_proc.wait()
        self._capture_proc = None

    def capture_frame(self, device: Camera, *args, **kwargs):
        from PIL import Image

        camera = self.prepare_device(device)
        frame = camera.stdout.read(
            device.info.resolution[0] * device.info.resolution[1] * 3
        )
        return Image.frombytes('RGB', device.info.resolution, frame)

    def to_grayscale(self, image):
        from PIL import Image

        new_image = Image.new('L', image.size)

        for i in range(0, image.size[0]):
            for j in range(0, image.size[1]):
                r, g, b = image.getpixel((i, j))
                value = int(2.0 * r - 1.125 * g - 1.75 * b)

                if value > 255:
                    value = 255
                if value < 0:
                    value = 0

                new_image.putpixel((i, j), value)

        return new_image

    @action
    def capture(self, output_file=None, *args, **kwargs):
        """
        Back-compatibility alias for :meth:`.capture_image`.
        """
        return self.capture_image(image_file=output_file, *args, **kwargs)


# vim:sw=4:ts=4:et:



import threading
from typing import Any, Optional, Dict, Union, Collection

from platypush.context import get_bus
from platypush.message.event.gpio import GPIOEvent
from platypush.plugins import RunnablePlugin, action


class GpioPlugin(RunnablePlugin):
    """
    This plugin can be used to interact with custom electronic devices
    connected to a Raspberry Pi (or compatible device) over GPIO pins.
    """

    def __init__(
        self,
        pins: Optional[Dict[str, int]] = None,
        monitored_pins: Optional[Collection[Union[str, int]]] = None,
        mode: str = 'board',
        **kwargs
    ):
        """
        :param mode: Specify ``board`` if you want to use the board PIN numbers,
            ``bcm`` for Broadcom PIN numbers (default: ``board``)
        :param pins: Custom `GPIO name` -> `PIN number` mapping. This can be
            useful if you want to reference your GPIO ports by name instead
            of PIN number.

        Example:

          .. code-block:: yaml

            pins:
                LED_1: 14,
                LED_2: 15,
                MOTOR: 16,
                SENSOR_1: 17
                SENSOR_2: 18

        :param monitored_pins: List of PINs to monitor. If a new value is detected
            on these pins then a :class:`platypush.message.event.gpio.GPIOEvent`
            event will be triggered. GPIO PINS can be referenced either by number
            or name, if a name is specified on the `pins` argument.
        """

        super().__init__(**kwargs)
        self.mode = self._get_mode(mode)
        self._initialized = False
        self._init_lock = threading.RLock()
        self._initialized_pins = {}
        self._monitored_pins = monitored_pins or []
        self.pins_by_name = pins if pins else {}
        self.pins_by_number = {
            number: name for (name, number) in self.pins_by_name.items()
        }

    def _init_board(self):
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized and GPIO.getmode():
                return

            GPIO.setmode(self.mode)
            self._initialized = True

    def _get_pin_number(self, pin):
        try:
            pin = int(str(pin))
        except ValueError:
            pin = self.pins_by_name.get(pin)
            if not pin:
                raise RuntimeError('Unknown PIN name: "{}"'.format(pin))

        return pin

    @staticmethod
    def _get_mode(mode_str: str) -> int:
        import RPi.GPIO as GPIO

        mode_str = mode_str.upper()
        assert mode_str in ['BOARD', 'BCM'], 'Invalid mode: {}'.format(mode_str)
        return getattr(GPIO, mode_str)

    def on_gpio_event(self):
        def callback(pin: int):
            import RPi.GPIO as GPIO

            value = GPIO.input(pin)
            pin = self.pins_by_number.get(pin, pin)
            get_bus().post(GPIOEvent(pin=pin, value=value))

        return callback

    def main(self):
        import RPi.GPIO as GPIO

        if not self._monitored_pins:
            return  # No need to start the monitor

        self._init_board()
        monitored_pins = [self._get_pin_number(pin) for pin in self._monitored_pins]

        for pin in monitored_pins:
            GPIO.setup(pin, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
            GPIO.add_event_detect(pin, GPIO.BOTH, callback=self.on_gpio_event())

        self.wait_stop()

    @action
    def write(
        self, pin: Union[int, str], value: Union[int, bool], name: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Write a byte value to a pin.

        :param pin: PIN number or configured name
        :param name: Optional name for the written value (e.g. "temperature" or "humidity")
        :param value: Value to write

        Response::

            output = {
                "name": <pin or metric name>,
                "pin": <pin>,
                "value": <value>,
                "method": "write"
            }
        """

        import RPi.GPIO as GPIO

        self._init_board()
        name = name or pin
        pin = self._get_pin_number(pin)

        if pin not in self._initialized_pins or self._initialized_pins[pin] != GPIO.OUT:
            GPIO.setup(pin, GPIO.OUT)
            self._initialized_pins[pin] = GPIO.OUT

        GPIO.setup(pin, GPIO.OUT)
        GPIO.output(pin, value)

        return {
            'name': name,
            'pin': pin,
            'value': value,
            'method': 'write',
        }

    @action
    def read(self, pin: Union[int, str], name: Optional[str] = None) -> Dict[str, Any]:
        """
        Reads a value from a PIN.

        :param pin: PIN number or configured name.
        :param name: Optional name for the read value (e.g. "temperature" or "humidity")

        Response::

            output = {
                "name": <pin number or pin/metric name>,
                "pin": <pin>,
                "value": <value>,
                "method": "read"
            }
        """

        import RPi.GPIO as GPIO

        self._init_board()
        name = name or pin
        pin = self._get_pin_number(pin)

        if pin not in self._initialized_pins:
            GPIO.setup(pin, GPIO.IN)
            self._initialized_pins[pin] = GPIO.IN

        val = GPIO.input(pin)

        return {
            'name': name,
            'pin': pin,
            'value': val,
            'method': 'read',
        }

    @action
    def get_measurement(self, pin=None):
        if pin is None:
            return self.read_all()
        return self.read(pin)

    @action
    def read_all(self):
        """
        Reads the values from all the configured PINs and returns them as a list. It will raise a RuntimeError if no
        PIN mappings were configured.
        """

        if not self.pins_by_number:
            raise RuntimeError("No PIN mappings were provided/configured")

        values = []
        for pin, name in self.pins_by_number.items():
            # noinspection PyUnresolvedReferences
            values.append(self.read(pin=pin, name=name).output)

        return values

    @action
    def cleanup(self):
        """
        Cleanup the state of the GPIO and resets PIN values.
        """
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized:
                GPIO.cleanup()
                self._initialized_pins = {}
                self._initialized = False


# vim:sw=4:ts=4:et:



import datetime
from typing import Dict, Optional

import requests
from dateutil.tz import gettz

from platypush.message.event.sun import SunriseEvent, SunsetEvent
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.sun import SunEventsSchema
from platypush.utils import utcnow


class SunPlugin(RunnablePlugin):
    """
    Plugin to get sunset/sunrise events and info for a certain location.
    """

    _base_url = 'https://api.sunrise-sunset.org/json'
    _schema = SunEventsSchema()
    _attr_to_event_class = {
        'sunrise': SunriseEvent,
        'sunset': SunsetEvent,
    }

    def __init__(self, latitude: float, longitude: float, **kwargs):
        """
        :param latitude: Default latitude.
        :param longitude: Default longitude.
        """
        super().__init__(**kwargs)
        self.latitude = latitude
        self.longitude = longitude

    def main(self):
        while not self.should_stop():
            next_events = self._get_events()
            next_event = next(
                iter(
                    sorted(
                        [
                            event_class(
                                latitude=self.latitude,
                                longitude=self.longitude,
                                time=next_events[attr],
                            )
                            for attr, event_class in self._attr_to_event_class.items()
                            if next_events.get(attr)
                        ],
                        key=lambda t: t.time,
                    )
                ),
                None,
            )

            assert next_event is not None, 'No next event found'
            wait_secs = max(
                0, (next_event.time - datetime.datetime.now(tz=gettz())).seconds
            )
            self.wait_stop(wait_secs)

            if not self.should_stop():
                self._bus.post(next_event)
                self.wait_stop(2)

    @staticmethod
    def _convert_time(t: str) -> datetime.datetime:
        now = utcnow().replace(microsecond=0)
        dt = datetime.datetime.strptime(
            f'{now.year}-{now.month:02d}-{now.day:02d} {t}',
            '%Y-%m-%d %I:%M:%S %p',
        ).replace(tzinfo=datetime.timezone.utc)

        if dt < now:
            dt += datetime.timedelta(days=1)
        return dt

    def _get_events(
        self, latitude: Optional[float] = None, longitude: Optional[float] = None
    ) -> Dict[str, datetime.datetime]:
        response = (
            requests.get(
                self._base_url,
                timeout=10,
                params={
                    'lat': latitude or self.latitude,
                    'lng': longitude or self.longitude,
                },
            )
            .json()
            .get('results', {})
        )

        return {
            attr: self._convert_time(t)
            for attr, t in response.items()
            if attr in self._schema.declared_fields
        }

    @action
    def get_events(
        self, latitude: Optional[float] = None, longitude: Optional[float] = None
    ) -> dict:
        """
        Return the next sun events.

        :param latitude: Override the default latitude.
        :param longitude: Override the default longitude.
        :return: .. schema:: sun.SunEventsSchema
        """
        schema = SunEventsSchema()
        return dict(
            schema.dump(self._get_events(latitude=latitude, longitude=longitude))
        )



import queue
import threading
from typing import Any, Collection, Dict, List, Optional, Tuple, Union

import requests

from platypush.entities import (
    DimmerEntityManager,
    EnumSwitchEntityManager,
    Entity,
    LightEntityManager,
    SwitchEntityManager,
)
from platypush.entities.devices import Device
from platypush.entities.dimmers import Dimmer
from platypush.entities.electricity import CurrentSensor, PowerSensor, VoltageSensor
from platypush.entities.lights import Light
from platypush.entities.humidity import HumiditySensor
from platypush.entities.motion import MotionSensor
from platypush.entities.sensors import BinarySensor, EnumSensor, NumericSensor
from platypush.entities.switches import EnumSwitch, Switch
from platypush.entities.temperature import TemperatureSensor
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.switchbot import DeviceSchema, DeviceStatusSchema, SceneSchema

from ._constants import DeviceType
from ._setters import entity_setters


# pylint: disable=too-many-ancestors
class SwitchbotPlugin(
    RunnablePlugin,
    DimmerEntityManager,
    EnumSwitchEntityManager,
    LightEntityManager,
    SwitchEntityManager,
):
    """
    Plugin to interact with the devices registered to a `Switchbot`_
    account/hub.

    In order to use this plugin:

        - Set up a Switchbot Hub and configure your devices through the
          Switchbot app.
        - Follow the steps on `get started page`_ to get an API token from the app.

    .. _get started page: https://github.com/OpenWonderLabs/SwitchBotAPI#getting-started
    .. _Switchbot: https://www.switch-bot.com/

    """

    def __init__(self, api_token: str, **kwargs):
        """
        :param api_token: API token (see `get started page`_).
        """
        super().__init__(**kwargs)
        self._api_token = api_token
        self._devices_by_id: Dict[str, dict] = {}
        self._devices_by_name: Dict[str, dict] = {}

    @staticmethod
    def _url_for(*args, device=None):
        url = 'https://api.switch-bot.com/v1.0/'
        if device:
            url += f'devices/{device["id"]}/'
        url += '/'.join(args)
        return url

    # pylint: disable=keyword-arg-before-vararg
    def _run(self, method: str = 'get', *args, device=None, **kwargs):
        response = getattr(requests, method)(
            self._url_for(*args, device=device),
            headers={
                'Authorization': self._api_token,
                'Accept': 'application/json',
                'Content-Type': 'application/json; charset=utf-8',
            },
            timeout=10,
            **kwargs,
        )

        response.raise_for_status()
        response = response.json()
        assert (
            response.get('statusCode') == 100
        ), f'Switchbot API request failed: {response.get("statusCode")}: {response.get("message")}'

        return response.get('body')

    @staticmethod
    def _split_device_id_and_property(device: str) -> Tuple[str, Optional[str]]:
        tokens = device.split(':')[:2]
        return tokens[0], (tokens[1] if len(tokens) == 2 else None)

    def _get_device(self, device: str, use_cache=True) -> dict:
        if not use_cache:
            self.devices()

        if device in self._devices_by_name:
            return self._devices_by_name[device]

        device, _ = self._split_device_id_and_property(device)
        if device in self._devices_by_id:
            return self._devices_by_id[device]

        assert use_cache, f'Device not found: {device}'
        return self._get_device(device, use_cache=False)

    @action
    def devices(self) -> List[dict]:
        """
        Get the list of devices associated to the specified Switchbot API account.

        :return: .. schema:: switchbot.DeviceSchema(many=True)
        """
        devices = self._run('get', 'devices')
        devices = [
            DeviceSchema().dump(
                {
                    **device,
                    'is_virtual': False,
                }
            )
            for device in devices.get('deviceList', [])
        ] + [
            DeviceSchema().dump(
                {
                    **device,
                    'is_virtual': True,
                }
            )
            for device in devices.get('infraredRemoteList', [])
        ]

        for device in devices:
            self._devices_by_id[device['id']] = device
            self._devices_by_name[device['name']] = device

        return devices

    @staticmethod
    def _get_device_metadata(device: dict) -> dict:
        return {
            "device_type": device.get("device_type"),
            "is_virtual": device.get("is_virtual", False),
            "hub_id": device.get("hub_id"),
        }

    @classmethod
    def _get_device_base(cls, device_dict: dict) -> Device:
        args: Dict[str, Any] = {
            'data': cls._get_device_metadata(device_dict),
        }

        return Device(
            id=f'{device_dict["id"]}',
            name=f'{device_dict["name"]}',
            **args,
        )

    @staticmethod
    def _matches_device_types(device: dict, *device_types: DeviceType) -> bool:
        return device.get('device_type') in {
            device_type.value for device_type in device_types
        }

    @classmethod
    def _get_bots(cls, *entities: dict) -> List[EnumSwitch]:
        return [
            EnumSwitch(
                id=dev["id"],
                name=dev["name"],
                value="on" if dev.get("on") else "off",
                values=["on", "off", "press"],
                is_write_only=True,
                data=cls._get_device_metadata(dev),
            )
            for dev in (entities or [])
            if cls._matches_device_types(dev, DeviceType.BOT)
        ]

    @classmethod
    def _get_lights(cls, *entities: dict) -> List[Light]:
        return [
            Light(
                id=dev["id"],
                name=dev["name"],
                on="on" if dev.get("on") else "off",
                brightness=dev.get("brightness"),
                color_temperature=dev.get("color_temperature"),
                color=dev.get("color"),
                data=cls._get_device_metadata(dev),
            )
            for dev in (entities or [])
            if cls._matches_device_types(
                dev,
                DeviceType.CEILING_LIGHT,
                DeviceType.CEILING_LIGHT_PRO,
                DeviceType.COLOR_BULB,
                DeviceType.STRIP_LIGHT,
            )
        ]

    @classmethod
    def _get_curtains(cls, *entities: dict) -> List[Dimmer]:
        return [
            Dimmer(
                id=dev["id"],
                name=dev["name"],
                value=dev.get("position"),
                min=0,
                max=100,
                unit='%',
                data=cls._get_device_metadata(dev),
            )
            for dev in (entities or [])
            if cls._matches_device_types(dev, DeviceType.CURTAIN)
        ]

    @classmethod
    def _get_meters(cls, device_dict: dict) -> List[Device]:
        devices = [cls._get_device_base(device_dict)]
        if device_dict.get('temperature') is not None:
            devices[0].children.append(
                TemperatureSensor(
                    id=f'{device_dict["id"]}:temperature',
                    name='Temperature',
                    value=device_dict['temperature'],
                    unit='C',
                )
            )

        if device_dict.get('humidity') is not None:
            devices[0].children.append(
                HumiditySensor(
                    id=f'{device_dict["id"]}:humidity',
                    name='Humidity',
                    value=device_dict['humidity'],
                    min=0,
                    max=100,
                    unit='%',
                )
            )

        if not devices[0].children:
            return []
        return devices

    @classmethod
    def _get_motion_sensors(cls, device_dict: dict) -> List[Device]:
        devices = [cls._get_device_base(device_dict)]
        if device_dict.get('moveDetected') is not None:
            devices[0].children.append(
                MotionSensor(
                    id=f'{device_dict["id"]}:motion',
                    name='Motion Detected',
                    value=bool(device_dict['moveDetected']),
                )
            )

        if device_dict.get('brightness') is not None:
            devices[0].children.append(
                BinarySensor(
                    id=f'{device_dict["id"]}:brightness',
                    name='Bright',
                    value=device_dict['brightness'] == 'bright',
                )
            )

        if not devices[0].children:
            return []
        return devices

    @classmethod
    def _get_contact_sensors(cls, device_dict: dict) -> List[Device]:
        devices = cls._get_motion_sensors(device_dict)
        if not devices:
            return []

        if device_dict.get('openState') is not None:
            devices[0].children.append(
                EnumSensor(
                    id=f'{device_dict["id"]}:open',
                    name='Open State',
                    value=device_dict['openState'],
                    values=['open', 'close', 'timeOutNotClose'],
                )
            )

        return devices

    @classmethod
    def _get_sensors(cls, *entities: dict) -> List[Device]:
        sensors: List[Entity] = []
        for dev in entities:
            if cls._matches_device_types(dev, DeviceType.METER, DeviceType.METER_PLUS):
                sensors.extend(cls._get_meters(dev))
            elif cls._matches_device_types(dev, DeviceType.MOTION_SENSOR):
                sensors.extend(cls._get_motion_sensors(dev))
            elif cls._matches_device_types(dev, DeviceType.CONTACT_SENSOR):
                sensors.extend(cls._get_contact_sensors(dev))

        return sensors

    @classmethod
    def _get_humidifiers(cls, *entities: dict) -> List[Device]:
        humidifiers = [
            dev
            for dev in entities
            if cls._matches_device_types(dev, DeviceType.HUMIDIFIER)
        ]

        devs = [Device(**cls._get_device_base(dev)) for dev in humidifiers]

        for dev_dict, entity in zip(humidifiers, devs):
            if dev_dict.get('power') is not None:
                entity.children.append(
                    Switch(
                        id=f'{dev_dict["id"]}:state',
                        name='State',
                        state=cls._is_on(dev_dict['power']),
                    )
                )

            if dev_dict.get('auto') is not None:
                entity.children.append(
                    Switch(
                        id=f'{dev_dict["id"]}:auto',
                        name='Automatic Mode',
                        state=cls._is_on(dev_dict['auto']),
                    )
                )

            if dev_dict.get('child_lock') is not None:
                entity.children.append(
                    Switch(
                        id=f'{dev_dict["id"]}:child_lock',
                        name='Child Lock',
                        state=cls._is_on(dev_dict['child_lock']),
                    )
                )

            if dev_dict.get('nebulization_efficiency') is not None:
                entity.children.append(
                    Dimmer(
                        id=f'{dev_dict["id"]}:nebulization_efficiency',
                        name='Nebulization Efficiency',
                        value=cls._is_on(dev_dict['nebulization_efficiency']),
                        min=0,
                        max=100,
                    )
                )

            if dev_dict.get('low_water') is not None:
                entity.children.append(
                    BinarySensor(
                        id=f'{dev_dict["id"]}:low_water',
                        name='Low Water',
                        value=cls._is_on(dev_dict['low_water']),
                    )
                )

            if dev_dict.get('temperature') is not None:
                entity.children.append(
                    TemperatureSensor(
                        id=f'{dev_dict["id"]}:temperature',
                        name='temperature',
                        value=dev_dict['temperature'],
                    )
                )

            if dev_dict.get('humidity') is not None:
                entity.children.append(
                    HumiditySensor(
                        id=f'{dev_dict["id"]}:humidity',
                        name='humidity',
                        value=dev_dict['humidity'],
                    )
                )

        return devs

    @classmethod
    def _get_locks(cls, *entities: dict) -> List[Device]:
        locks = [
            dev
            for dev in (entities or [])
            if cls._matches_device_types(dev, DeviceType.LOCK)
        ]

        devices = [Device(**cls._get_device_base(plug)) for plug in locks]

        for plug, device in zip(locks, devices):
            if plug.get('locked') is not None:
                device.children.append(
                    Switch(
                        id=f'{plug["id"]}:locked',
                        name='Locked',
                        state=cls._is_on(plug['locked']),
                    )
                )

            if plug.get('door_open') is not None:
                device.children.append(
                    BinarySensor(
                        id=f'{plug["id"]}:door_open',
                        name='Door Open',
                        value=cls._is_on(plug['door_open']),
                    )
                )

        return devices

    @classmethod
    def _get_plugs(cls, *entities: dict) -> List[Device]:
        plugs = [
            dev
            for dev in (entities or [])
            if cls._matches_device_types(
                dev, DeviceType.PLUG, DeviceType.PLUG_MINI_JP, DeviceType.PLUG_MINI_US
            )
        ]

        devices = [Device(**cls._get_device_base(plug)) for plug in plugs]

        for plug, device in zip(plugs, devices):
            if plug.get('on') is not None:
                device.children.append(
                    Switch(
                        id=f'{plug["id"]}:state',
                        name='State',
                        state=cls._is_on(plug['on']),
                    )
                )

            if plug.get('power') is not None:
                device.children.append(
                    PowerSensor(
                        id=f'{plug["id"]}:power',
                        name='Power',
                        value=plug['power'],
                        unit='W',
                    )
                )

            if plug.get('voltage') is not None:
                device.children.append(
                    VoltageSensor(
                        id=f'{plug["id"]}:voltage',
                        name='Voltage',
                        value=plug['voltage'],
                        unit='V',
                    )
                )

            if plug.get('current') is not None:
                device.children.append(
                    CurrentSensor(
                        id=f'{plug["id"]}:current',
                        name='Current',
                        value=plug['current'],
                        unit='A',
                    )
                )

            if plug.get('active_time') is not None:
                device.children.append(
                    NumericSensor(
                        id=f'{plug["id"]}:active_time',
                        name='Active Time',
                        value=plug['active_time'],
                        unit='min',
                    )
                )

        return devices

    @staticmethod
    def _is_on(state: Union[bool, str, int]) -> bool:
        if isinstance(state, str):
            state = state.lower()
        else:
            state = bool(state)
        return state in {'on', 'true', '1', True}

    def transform_entities(self, entities: Collection[dict]) -> Collection[Entity]:
        return [
            *self._get_bots(*entities),
            *self._get_curtains(*entities),
            *self._get_humidifiers(*entities),
            *self._get_lights(*entities),
            *self._get_locks(*entities),
            *self._get_plugs(*entities),
            *self._get_sensors(*entities),
        ]

    def _worker(  # pylint: disable=keyword-arg-before-vararg
        self,
        q: queue.Queue,
        method: str = 'get',
        *args,
        device: Optional[dict] = None,
        **kwargs,
    ):
        schema = DeviceStatusSchema()
        try:
            if (
                method == 'get'
                and args
                and args[0] == 'status'
                and device
                and device.get('is_virtual')
            ):
                res = schema.load(device)
            else:
                res = self._run(method, *args, device=device, **kwargs)

            q.put(schema.dump(res))
        except Exception as e:
            self.logger.exception(e)
            q.put(e)

    @action
    # pylint: disable=arguments-differ
    def status(
        self, device: Optional[str] = None, publish_entities: bool = True, **_
    ) -> Union[dict, List[dict]]:
        """
        Get the status of all the registered devices or of a specific device.

        :param device: Filter by device ID or name.
        :return: .. schema:: switchbot.DeviceStatusSchema(many=True)
        """
        devices = self.devices().output
        if device:
            device_info = self._get_device(device)
            status = (
                {}
                if device_info['is_virtual']
                else self._run('get', 'status', device=device_info)
            )
            return {
                **device_info,
                **status,
            }

        devices_by_id = {dev['id']: dev for dev in devices}
        queues: List[queue.Queue] = [queue.Queue()] * len(devices)
        workers = [
            threading.Thread(
                target=self._worker,
                args=(queues[i], 'get', 'status'),
                kwargs={'device': dev},
            )
            for i, dev in enumerate(devices)
        ]

        results = []
        for worker in workers:
            worker.start()

        for q in queues:
            response = q.get()
            if not response:
                continue

            assert not isinstance(response, Exception), str(response)
            results.append(
                {
                    **devices_by_id.get(response.get('id'), {}),
                    **response,
                }
            )

        for worker in workers:
            worker.join()

        if publish_entities:
            self.publish_entities(results)
        return results

    @action
    def press(self, device: str):
        """
        Send a press-button command to a device.

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run('post', 'commands', device=dev, json={'command': 'press'})

    @action
    def toggle(self, device: str, **_):  # pylint: disable=arguments-differ
        """
        Shortcut for :meth:`.press`.

        :param device: Device name or ID.
        """
        return self.press(device)

    @action
    def on(self, device: str, **_):  # pylint: disable=arguments-differ
        """
        Send a turn-on command to a device

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run('post', 'commands', device=dev, json={'command': 'turnOn'})

    @action
    def off(self, device: str, **_):  # pylint: disable=arguments-differ
        """
        Send a turn-off command to a device

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run('post', 'commands', device=dev, json={'command': 'turnOff'})

    @action
    def lock(self, device: str, **_):
        """
        Lock a compatible lock device.

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run('post', 'commands', device=dev, json={'command': 'lock'})

    @action
    def unlock(self, device: str, **_):
        """
        Unlock a compatible lock device.

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run('post', 'commands', device=dev, json={'command': 'unlock'})

    @action
    def set_curtain_position(self, device: str, position: int):
        """
        Set the position of a curtain device.

        :param device: Device name or ID.
        :param position: An integer between 0 (open) and 100 (closed).
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'setPosition',
                'commandType': 'command',
                'parameter': f'0,ff,{position}',
            },
        )

    @action
    def set_humidifier_efficiency(self, device: str, efficiency: Union[int, str]):
        """
        Set the nebulization efficiency of a humidifier device.

        :param device: Device name or ID.
        :param efficiency: Possible values:

            - ``auto``: Automatic mode.
            - A value between ``0`` and ``100``.

        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'setMode',
                'commandType': 'command',
                'parameter': efficiency,
            },
        )

    @action
    def set_fan_speed(self, device: str, speed: int):
        """
        Set the speed of a fan.

        :param device: Device name or ID.
        :param speed: Speed between 1 and 4.
        """
        status = self.status(device=device).output
        mode = status.get('mode')
        swing_range = status.get('swing_range')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'set',
                'commandType': 'command',
                'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
            },
        )

    @action
    def set_fan_mode(self, device: str, mode: int):
        """
        Set the mode of a fan.

        :param device: Device name or ID.
        :param mode: Fan mode (1 or 2).
        """
        status = self.status(device=device).output
        speed = status.get('speed')
        swing_range = status.get('swing_range')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'set',
                'commandType': 'command',
                'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
            },
        )

    @action
    def set_swing_range(self, device: str, swing_range: int):
        """
        Set the swing range of a fan.

        :param device: Device name or ID.
        :param swing_range: Swing range angle, between 0 and 120.
        """
        status = self.status(device=device).output
        speed = status.get('speed')
        mode = status.get('mode')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'set',
                'commandType': 'command',
                'parameter': ','.join(['on', str(mode), str(speed), str(swing_range)]),
            },
        )

    @action
    def set_temperature(self, device: str, temperature: float):
        """
        Set the temperature of an air conditioner.

        :param device: Device name or ID.
        :param temperature: Temperature, in Celsius.
        """
        status = self.status(device=device).output
        mode = status.get('mode')
        fan_speed = status.get('fan_speed')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'setAll',
                'commandType': 'command',
                'parameter': ','.join(
                    [str(temperature), str(mode), str(fan_speed), 'on']
                ),
            },
        )

    @action
    def set_ac_mode(self, device: str, mode: int):
        """
        Set the mode of an air conditioner.

        :param device: Device name or ID.
        :param mode: Air conditioner mode. Supported values:

            * 1: ``auto``
            * 2: ``cool``
            * 3: ``dry``
            * 4: ``fan``
            * 5: ``heat``

        """
        status = self.status(device=device).output
        temperature = status.get('temperature')
        fan_speed = status.get('fan_speed')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'setAll',
                'commandType': 'command',
                'parameter': ','.join(
                    [str(temperature), str(mode), str(fan_speed), 'on']
                ),
            },
        )

    @action
    def set_ac_fan_speed(self, device: str, fan_speed: int):
        """
        Set the fan speed for an air conditioner.

        :param device: Device name or ID.
        :param fan_speed: Possible values:

            * 1: ``auto``
            * 2: ``low``
            * 3: ``medium``
            * 4: ``high``

        """
        status = self.status(device=device).output
        temperature = status.get('temperature')
        mode = status.get('mode')
        return self._run(
            'post',
            'commands',
            device=device,
            json={
                'command': 'setAll',
                'commandType': 'command',
                'parameter': ','.join(
                    [str(temperature), str(mode), str(fan_speed), 'on']
                ),
            },
        )

    @action
    def set_channel(self, device: str, channel: int):
        """
        Set the channel on a TV, IPTV/Streamer, Set Top Box device.

        :param device: Device name or ID.
        :param channel: Channel number.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'SetChannel',
                'commandType': 'command',
                'parameter': [str(channel)],
            },
        )

    @action
    def volup(self, device: str):
        """
        Send volume up IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'volumeAdd',
                'commandType': 'command',
            },
        )

    @action
    def voldown(self, device: str):
        """
        Send volume down IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'volumeSub',
                'commandType': 'command',
            },
        )

    @action
    def mute(self, device: str):
        """
        Send mute/unmute IR event to a device (for TV, IPTV/Streamer, Set Top Box, DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'setMute',
                'commandType': 'command',
            },
        )

    @action
    def channel_next(self, device: str):
        """
        Send next channel IR event to a device (for TV, IPTV/Streamer, and Set Top Box).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'channelAdd',
                'commandType': 'command',
            },
        )

    @action
    def channel_prev(self, device: str):
        """
        Send previous channel IR event to a device (for TV, IPTV/Streamer, and Set Top Box).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'channelSub',
                'commandType': 'command',
            },
        )

    @action
    def play(self, device: str):
        """
        Send play IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Play',
                'commandType': 'command',
            },
        )

    @action
    def pause(self, device: str):
        """
        Send pause IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Pause',
                'commandType': 'command',
            },
        )

    @action
    def ir_stop(self, device: str):
        """
        Send stop IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Stop',
                'commandType': 'command',
            },
        )

    @action
    def forward(self, device: str):
        """
        Send forward IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'FastForward',
                'commandType': 'command',
            },
        )

    @action
    def back(self, device: str):
        """
        Send backward IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Rewind',
                'commandType': 'command',
            },
        )

    @action
    def next(self, device: str):
        """
        Send next IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Next',
                'commandType': 'command',
            },
        )

    @action
    def previous(self, device: str):
        """
        Send previous IR event to a device (for DVD and Speaker).

        :param device: Device name or ID.
        """
        dev = self._get_device(device)
        return self._run(
            'post',
            'commands',
            device=dev,
            json={
                'command': 'Previous',
                'commandType': 'command',
            },
        )

    @action
    def scenes(self) -> List[dict]:
        """
        Get the list of registered scenes.

        :return: .. schema:: switchbot.SceneSchema(many=True)
        """
        return SceneSchema().dump(self._run('get', 'scenes'), many=True)

    @action
    def run_scene(self, scene: str):
        """
        Execute a scene.

        :param scene: Scene ID or name.
        """
        scenes = [
            s
            for s in self.scenes().output
            if s.get('id') == scene or s.get('name') == scene
        ]

        assert scenes, f'No such scene: {scene}'
        return self._run('post', 'scenes', scenes[0]['id'], 'execute')

    @action
    # pylint: disable=redefined-builtin
    def set_value(
        self, device: str, property: Optional[str] = None, value: Any = None, **__
    ):
        """
        Set the value of a property of a device.

        :param device: Device name or ID, or entity (external) ID.
        :param property: Property to set. It should be present if you are
            passing a root device ID to ``device`` and not an atomic entity in
            the format ``<device_id>:<property_name>``.
        :param value: Value to set.
        """
        entity = self._to_entity(device, property)
        assert entity, f'No such device: "{device}"'

        dt = entity.data.get('device_type')
        assert dt, f'Could not infer the device type for "{device}"'

        device_type = DeviceType(dt)
        setter_class = entity_setters.get(device_type)
        assert setter_class, f'No setters found for device type "{device_type}"'

        setter = setter_class(entity)
        return setter(property=property, value=value)

    @action
    def set(self, entity: str, value: Any, attribute: Optional[str] = None, **kwargs):
        return self.set_value(entity, property=attribute, value=value, **kwargs)

    def _to_entity(
        self,
        device: str,
        property: Optional[str] = None,  # pylint: disable=redefined-builtin
    ) -> Optional[Entity]:
        dev = self._get_device(device)
        entities = list(self.transform_entities([dev]))
        if not entities:
            return None
        if len(entities) == 1:
            return entities[0]
        if not property:
            device, property = self._split_device_id_and_property(device)
        assert property, 'No property specified'

        entity_id = f'{device}:{property}'
        return next(iter([e for e in entities if e.id == entity_id]), None)

    @action
    def set_lights(
        self,
        *_,
        lights: Collection[str],
        on: Optional[bool] = None,
        brightness: Optional[int] = None,
        hex: Optional[str] = None,  # pylint: disable=redefined-builtin
        temperature: Optional[int] = None,
        **__,
    ):
        """
        Change the settings for compatible lights.

        :param lights: Light names or IDs.
        :param on: Turn on the lights.
        :param brightness: Set the brightness of the lights.
        :param hex: Set the color of the lights.
        :param temperature: Set the temperature of the lights.
        """
        devices = [self._get_device(light) for light in lights]
        for dev in devices:
            if on is not None:
                method = self.on if on else self.off
                method(dev['id'])

            if brightness is not None:
                self._run(
                    'post',
                    'commands',
                    device=dev,
                    json={
                        'command': 'setBrightness',
                        'commandType': 'command',
                        'parameter': brightness,
                    },
                )

            if hex is not None:
                self._run(
                    'post',
                    'commands',
                    device=dev,
                    json={
                        'command': 'setColor',
                        'commandType': 'command',
                        'parameter': hex,
                    },
                )

            if temperature is not None:
                self._run(
                    'post',
                    'commands',
                    device=dev,
                    json={
                        'command': 'setColorTemperature',
                        'commandType': 'command',
                        'parameter': temperature,
                    },
                )

    def main(self):
        entities = {}

        while not self.should_stop():
            status = self.status(publish_entities=False).output
            new_entities = {e['id']: e for e in status}
            updated_entities = {
                id: e
                for id, e in new_entities.items()
                if any(v != entities.get(id, {}).get(k) for k, v in e.items())
            }

            if updated_entities:
                self.publish_entities(updated_entities.values())

            entities = new_entities
            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



# mypy stub for ServiceClass

from enum import Enum

from platypush.plugins.bluetooth._types import RawServiceClass

class ServiceClass(Enum):
    """
    Enumeration of supported Bluetooth service classes.
    """

    value: RawServiceClass
    """ The raw service class value. """

    UNKNOWN = ...
    """ A class for unknown services. """
    OBEX_OBJECT_PUSH = ...
    """ Class for the OBEX Object Push service. """
    OBEX_FILE_TRANSFER = ...
    """ Class for the OBEX File Transfer service. """
    SWITCHBOT = ...
    """ Class for Switchbot devices services. """

    @classmethod
    def get(cls, value: RawServiceClass) -> "ServiceClass":
        """
        :param value: The raw service class UUID.
        :return: The parsed :class:`ServiceClass` instance, or
            ``ServiceClass.UNKNOWN``.
        """
    @classmethod
    def by_name(cls, name: str) -> "ServiceClass":
        """
        :param name: The name of the service class.
        :return: The :class:`ServiceClass` instance, or
            ``ServiceClass.UNKNOWN``.
        """



from abc import ABC, abstractmethod
from typing import Optional

from .._state import SerializedState, StateSerializer, XmppState
from ._base import XmppBaseMixin


# pylint: disable=too-few-public-methods
class XmppStateMixin(XmppBaseMixin, ABC):
    """
    A simple mixin that encapsulates an XMPP state object.
    """

    @abstractmethod
    def __init__(
        self,
        *args,
        state: Optional[XmppState] = None,
        loaded_state: Optional[SerializedState] = None,
        state_serializer: Optional[StateSerializer] = None,
        **kwargs,
    ):
        self._state = state or XmppState()
        self._loaded_state = loaded_state or SerializedState()
        self._state_serializer = state_serializer
        super().__init__(*args, **kwargs)

    def restore_state(self):
        """
        Function called by the plugin once connected to notify that the
        component should reload the previous state (optional, to be implemented
        by derived classes).
        """



import datetime
from typing import Any, Callable, Optional

import pychromecast  # type: ignore

from platypush.context import get_bus
from platypush.message.event.media import MediaEvent

MediaCallback = Callable[[MediaEvent, pychromecast.Chromecast], Optional[Any]]


def convert_status(status) -> dict:
    """
    Convert a Chromecast status object to a dictionary.
    """
    attrs = [
        a
        for a in dir(status)
        if not a.startswith('_') and not callable(getattr(status, a))
    ]

    renamed_attrs = {
        'current_time': 'position',
        'player_state': 'state',
        'supports_seek': 'seekable',
        'volume_level': 'volume',
        'volume_muted': 'muted',
        'content_id': 'url',
    }

    ret = {}
    for attr in attrs:
        value = getattr(status, attr)
        if attr == 'volume_level':
            value = round(100 * value, 2)
        if attr == 'player_state':
            value = value.lower()
            if value == 'paused':
                value = 'pause'
            if value == 'playing':
                value = 'play'
        if isinstance(value, datetime.datetime):
            value = value.isoformat()

        if attr in renamed_attrs:
            ret[renamed_attrs[attr]] = value
        else:
            ret[attr] = value

    return ret


def post_event(
    evt_type,
    callback: Optional[MediaCallback] = None,
    chromecast: Optional[pychromecast.Chromecast] = None,
    **evt
):
    evt['plugin'] = 'media.chromecast'
    event = evt_type(player=evt.get('device'), **evt)

    get_bus().post(event)
    if callback:
        callback(event, chromecast)



from typing import Optional

from platypush.backend.http.app.utils import logger
from platypush.backend.http.media.handlers import MediaHandler

from ._registry import load_media_map, save_media_map, media_map_lock


def unregister_media(source: Optional[str] = None):
    """
    Unregisters a media streaming URL file given its source.
    """
    assert source is not None, 'No media_id specified'
    media_id = MediaHandler.get_media_id(source)
    media_info = {}

    with media_map_lock:
        media_map = load_media_map()
        if media_id not in media_map:
            raise FileNotFoundError(f'{source} is not a registered media_id')

        media_info = media_map.pop(media_id)
        save_media_map(media_map)

    logger().info('Unregistered %s from %s', source, media_info.url)
    return media_info



from marshmallow import fields
from marshmallow.schema import Schema

from . import DateTime


class GotifyMessageSchema(Schema):
    title = fields.String(
        metadata=dict(
            description='Message title',
            example='Test title',
        )
    )

    message = fields.String(
        required=True,
        metadata=dict(
            description='Message body (markdown is supported)',
            example='Test message',
        )
    )

    priority = fields.Int(
        missing=0,
        metadata=dict(
            description='Message priority',
            example=2,
        )
    )

    extras = fields.Dict(
        metadata=dict(
            description='Extra payload to be delivered with the message',
            example={
                'home::appliances::lighting::on': {
                    'brightness': 15
                },
                'home::appliances::thermostat::change_temperature': {
                    'temperature': 23
                }
            },
        )
    )

    id = fields.Int(
        required=True,
        dump_only=True,
        metadata=dict(
            description='Message ID',
            example=1,
        )
    )

    appid = fields.Int(
        dump_only=True,
        metadata=dict(
            description='ID of the app that posted the message',
            example=1,
        )
    )

    date = DateTime(
        dump_only=True,
        metadata=dict(
            description='Message date',
        )
    )



from dataclasses import dataclass, field
from threading import Event, RLock
from typing import Optional


@dataclass
class PlaylistSync:
    """
    Object used to synchronize playlist load/change events between threads.
    """

    _loading_lock: RLock = field(default_factory=RLock)
    _loading: Event = field(default_factory=Event)
    _loaded: Event = field(default_factory=Event)

    def wait_for_loading(self, timeout: Optional[float] = None):
        """
        Wait for the playlist to be loaded.

        :param timeout: The maximum time to wait for the playlist to be loaded.
        """
        # If the loading event is not set, no playlist change - we can proceed
        # with notifying the event.
        if not self._loading.is_set():
            return True

        # Wait for the full playlist to be loaded.
        return self._loaded.wait(timeout)

    def __enter__(self):
        """
        Called when entering a context manager to handle a playlist loading
        session.
        """
        self._loading_lock.acquire()
        self._loading.set()
        self._loaded.clear()

    def __exit__(self, *_):
        """
        Called when exiting a context manager to handle a playlist loading
        session.
        """
        self._loading.clear()
        self._loaded.set()

        try:
            self._loading_lock.release()
        except RuntimeError:
            pass



import logging
import inspect
import json
import os
import pathlib
import subprocess
import sys
import types
from datetime import datetime
from typing import Callable, Dict, List, Optional, Set, Type, Tuple, Any

import pkgutil

from dateutil.tz import tzutc
from sqlalchemy import (
    Boolean,
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    JSON,
    String,
    UniqueConstraint,
    inspect as schema_inspect,
)
from sqlalchemy.engine import Engine
from sqlalchemy.orm import ColumnProperty, backref, relationship
from sqlalchemy.orm.exc import ObjectDeletedError

import platypush
from platypush.config import Config
from platypush.common.db import Base, is_defined
from platypush.message import JSONAble, Message
from platypush.utils import utcnow

EntityRegistryType = Dict[str, Type['Entity']]
entities_registry: EntityRegistryType = {}

EntityKey = Tuple[str, str]
""" The entity's logical key, as an ``<external_id, plugin>`` tuple. """
EntityMapping = Dict[EntityKey, 'Entity']
""" Internal mapping for entities used for deduplication/merge/upsert. """

_import_error_ignored_modules: Set[str] = {'bluetooth'}
"""
ImportError exceptions will be ignored for these entity submodules when
imported dynamically. An ImportError for these modules means that some optional
requirements are missing, and if those plugins aren't enabled then we shouldn't
fail.
"""

logger = logging.getLogger(__name__)


if not is_defined('entity'):

    class Entity(Base):
        """
        Model for a general-purpose platform entity.
        """

        __tablename__ = 'entity'

        id = Column(Integer, autoincrement=True, primary_key=True)
        external_id = Column(String, nullable=False)
        name = Column(String, nullable=False, index=True)
        description = Column(String)
        type = Column(String, nullable=False, index=True)
        plugin = Column(String, nullable=False)
        parent_id = Column(
            Integer,
            ForeignKey(f'{__tablename__}.id', ondelete='CASCADE'),
            nullable=True,
        )

        data = Column(JSON, default=dict)
        meta = Column(JSON, default=dict)
        is_read_only = Column(Boolean, default=False)
        is_write_only = Column(Boolean, default=False)
        is_query_disabled = Column(Boolean, default=False)
        is_configuration = Column(Boolean, default=False)
        external_url = Column(String)
        image_url = Column(String)

        created_at = Column(DateTime(timezone=False), default=utcnow(), nullable=False)
        updated_at = Column(
            DateTime(timezone=False),
            default=utcnow(),
            onupdate=utcnow(),
        )

        parent = relationship(
            'Entity',
            remote_side=[id],
            uselist=False,
            lazy='joined',
            post_update=True,
            backref=backref(
                'children',
                remote_side=[parent_id],
                uselist=True,
                cascade='all, delete-orphan',
                lazy='immediate',
            ),
        )

        UniqueConstraint(external_id, plugin)

        __table_args__ = (
            Index('name_and_plugin_index', name, plugin),
            Index('name_type_and_plugin_index', name, type, plugin),
            {'extend_existing': True},
        )

        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
            'polymorphic_on': type,
        }

        @classmethod
        def get_columns(cls) -> Tuple[ColumnProperty, ...]:
            inspector = schema_inspect(cls)
            return tuple(inspector.mapper.column_attrs)

        @property
        def entity_key(self) -> EntityKey:
            """
            This method returns the "external" key of an entity.
            """
            return str(self.external_id), str(self.plugin)

        def copy(self) -> 'Entity':
            """
            This method returns a copy of the entity. It's useful when you want
            to reuse entity objects in other threads or outside of their
            associated SQLAlchemy session context.
            """

            def key_value_pair(col: ColumnProperty):
                try:
                    return (col.key, getattr(self, col.key, None))
                except ObjectDeletedError:
                    return None

            return self.__class__(
                **dict(
                    key_value_pair(col)  # type: ignore
                    for col in self.get_columns()
                    if key_value_pair(col) is not None
                ),
                children=[child.copy() for child in self.children],
            )

        def _serialize_value(self, column_name: str) -> Any:
            val = getattr(self, column_name)
            if isinstance(val, datetime):
                # All entity timestamps are in UTC
                val = val.replace(tzinfo=tzutc()).isoformat()

            return val

        def _column_name(self, col: ColumnProperty) -> Optional[str]:
            """
            Normalizes the column name, taking into account native columns and
            columns mapped to properties.
            """
            try:
                normalized_name = col.key.lstrip('_')
                if len(col.key.lstrip('_')) == col.key or not hasattr(
                    self, normalized_name
                ):
                    return col.key  # It's not a hidden column with a mapped property

                return normalized_name
            except ObjectDeletedError as e:
                logger.warning(
                    'Could not access column "%s" for entity ID "{%s}": {%s}',
                    col.key,
                    self.id,
                    e,
                )
                return None

        def _column_to_pair(self, col: ColumnProperty) -> tuple:
            """
            Utility method that, given a column, returns a pair containing its
            normalized name and its serialized value.
            """
            col_name = self._column_name(col)
            if col_name is None:
                return ()
            return col_name, self._serialize_value(col_name)

        @property
        def children_ids(self) -> List[int]:
            """
            Returns the children IDs of the entity.
            """
            children_ids = []
            for child in self.children:
                try:
                    children_ids.append(child.id)
                except ObjectDeletedError:
                    pass

            return children_ids

        def to_dict(self) -> dict:
            """
            Returns the current entity as a flatten dictionary.
            """
            return {
                **dict(
                    self._column_to_pair(col)
                    for col in self.get_columns()
                    if self._column_to_pair(col)
                ),
                'children_ids': self.children_ids,
            }

        def to_json(self) -> dict:
            """
            Alias for :meth:`.to_dict`.
            """
            return self.to_dict()

        def __repr__(self):
            """
            Same as :meth:`.__str__`.
            """
            return str(self)

        def __str__(self):
            """
            :return: A JSON-encoded representation of the entity.
            """
            return json.dumps(self.to_dict(), cls=Message.Encoder)

        def __setattr__(self, key, value):
            """
            Serializes the new value before assigning it to an attribute.
            """
            matching_columns = [c for c in self.get_columns() if c.expression.name == key]  # type: ignore

            if (
                matching_columns
                and issubclass(type(matching_columns[0].columns[0].type), DateTime)
                and isinstance(value, str)
            ):
                value = datetime.fromisoformat(value)

            return super().__setattr__(key, value)

        def get_plugin(self):
            from platypush.context import get_plugin

            plugin = get_plugin(self.plugin)
            assert plugin, f'No such plugin: {plugin}'
            return plugin

        def run(self, action: str, *args, **kwargs):
            plugin = self.get_plugin()
            method = getattr(plugin, action, None)
            assert method, f'No such action: {self.plugin}.{action}'
            return method(self.external_id or self.name, *args, **kwargs)

    # Inject the JSONAble mixin (Python goes nuts if done through
    # standard multiple inheritance with an SQLAlchemy ORM class)
    Entity.__bases__ = Entity.__bases__ + (JSONAble,)

    EntitySavedCallback = Callable[[Entity], Any]
    """
    Type for the callback functions that should be called when an entity is saved
    on the database.
    """


def _discover_entity_types():
    for loader, modname, _ in pkgutil.walk_packages(
        path=[str(pathlib.Path(__file__).parent.absolute())],
        prefix=__package__ + '.',
        onerror=lambda _: None,
    ):
        try:
            mod_loader = loader.find_spec(modname, None)
            assert mod_loader and mod_loader.loader
            module = types.ModuleType(mod_loader.name)
            mod_loader.loader.exec_module(module)
        except Exception as e:
            if (
                isinstance(e, (ImportError, ModuleNotFoundError))
                and modname[len(__package__) + 1 :] in _import_error_ignored_modules
            ):
                logger.debug('Could not import module %s', modname)
            else:
                logger.warning('Could not import module %s', modname)
                logger.exception(e)

            continue

        for _, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, Entity):
                entities_registry[obj] = {}  # type: ignore


def _get_db():
    """
    Utility method to get the db plugin.
    """
    from platypush.context import get_plugin

    db = get_plugin('db')
    assert db
    return db


def _get_db_engine() -> Engine:
    """
    Utility method to get the db engine.
    """
    return _get_db().get_engine()


def get_entities_registry() -> EntityRegistryType:
    """
    :returns: A copy of the entities registry.
    """
    return entities_registry.copy()


def init_entities_db():
    """
    Initializes the entities database.
    """
    run_db_migrations()
    _discover_entity_types()
    _get_db().create_all(_get_db_engine(), Base)


def run_db_migrations():
    """
    Run the database migrations upon engine initialization.
    """
    logger.info('Running database migrations')
    alembic_ini = os.path.join(
        os.path.dirname(inspect.getabsfile(platypush)), 'migrations', 'alembic.ini'
    )

    subprocess.run(
        [
            sys.executable,
            '-m',
            'alembic',
            '-c',
            alembic_ini,
            '-x',
            f'CFGFILE={Config.get_file()}',
            '-x',
            f'DBNAME={_get_db_engine().url}',
            'upgrade',
            'head',
        ],
        check=True,
    )



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .sensors import NumericSensor


if not is_defined('power_sensor'):

    class PowerSensor(NumericSensor):
        __tablename__ = 'power_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('current_sensor'):

    class CurrentSensor(NumericSensor):
        __tablename__ = 'current_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('voltage_sensor'):

    class VoltageSensor(NumericSensor):
        __tablename__ = 'voltage_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('energy_sensor'):

    class EnergySensor(NumericSensor):
        __tablename__ = 'energy_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from typing import Collection

import dateutil.parser
import importlib

from abc import ABCMeta, abstractmethod

from platypush.context import get_plugin
from platypush.plugins import Plugin, action


class CalendarInterface:
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_upcoming_events(self, max_results=10):
        raise NotImplementedError()


class CalendarPlugin(Plugin, CalendarInterface):
    """
    The CalendarPlugin allows you to keep track of multiple calendars (Google or
    iCal URLs) and get joined events from all of them.
    """

    def __init__(self, calendars: Collection[dict], *args, **kwargs):
        """
        :param calendars: List of calendars to be queried. Supported types so far: Google Calendar and iCal URLs.

        Example:

            .. code-block:: yaml

                calendars:
                    # Use the Google Calendar integration
                    - type: google.calendar

                    # Import the Facebook events calendar via iCal URL
                    - type: calendar.ical
                      url: https://www.facebook.com/ical/u.php?uid=USER_ID&key=FB_KEY

        """

        super().__init__(*args, **kwargs)
        self.calendars = []

        for calendar in calendars:
            cal_type = calendar.pop('type', None)
            if cal_type is None:
                self.logger.warning(
                    "Invalid calendar with no type specified: %s", calendar
                )
                continue

            try:
                # New `calendar.name` format
                cal_plugin = get_plugin(cal_type).__class__
            except Exception:
                # Legacy `platypush.plugins.calendar.name.CalendarNamePlugin` format
                module_name = '.'.join(cal_type.split('.')[:-1])
                class_name = cal_type.split('.')[-1]
                module = importlib.import_module(module_name)
                cal_plugin = getattr(module, class_name)

            self.calendars.append(cal_plugin(**calendar))

    @action
    def get_upcoming_events(self, max_results=10):
        """
        Get a list of upcoming events merging all the available calendars.

        :param max_results: Maximum number of results to be returned (default: 10)
        :type max_results: int

        :returns: platypush.message.Response -- Response object with the list of
            events in the Google calendar API format.

        Example::

            output = [
                {
                    "id": "123456abcdef",
                    "kind": "calendar#event",
                    "status": "confirmed",
                    "htmlLink": "...",
                    "created": "2018-06-01T01:23:45.000Z",
                    "updated": "2018-06-01T01:23:45.000Z",
                    "creator": {
                        "email": "...",
                        "displayName": "...",
                        "self": true
                    },
                    "organizer" {
                        "email": "...",
                        "displayName": "...",
                        "self": true
                    },
                    "start": {
                        "dateTime": "2018-06-02T10:00:00.000Z",
                    },
                    "end": {
                        "dateTime": "2018-06-02T12:00:00.000Z",
                    },
                },
                ...
            ]
        """

        events = []

        for calendar in self.calendars:
            try:
                cal_events = calendar.get_upcoming_events().output or []
                events.extend(cal_events)
            except Exception as e:
                self.logger.warning(
                    'Could not retrieve events from calendar %s: %s', calendar, e
                )

        events = sorted(
            events,
            key=lambda event: dateutil.parser.parse(
                event['start']['dateTime']
                if 'dateTime' in event['start']
                else event['start']['date'] + 'T00:00:00+00:00'
            ),
        )[:max_results]

        return events


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.backend.http.app.utils import logger, send_request
from platypush.backend.http.media.handlers import MediaHandler

from ._registry import load_media_map, save_media_map


def get_media_url(media_id: str) -> str:
    """
    :returns: The URL of a media file given its ID
    """
    return f'/media/{media_id}'


def register_media(source: str, subtitles: Optional[str] = None) -> MediaHandler:
    """
    Registers a media file and returns its associated media handler.
    """
    media_id = MediaHandler.get_media_id(source)
    media_url = get_media_url(media_id)
    media_map = load_media_map()
    subfile = None

    if subtitles:
        req = {
            'type': 'request',
            'action': 'media.subtitles.download',
            'args': {
                'link': subtitles,
                'convert_to_vtt': True,
            },
        }

        try:
            subfile = (send_request(req) or {}).get('filename')
        except Exception as e:
            logger().warning('Unable to load subtitle %s: %s', subtitles, e)

    media_hndl = MediaHandler.build(source, url=media_url, subtitles=subfile)
    media_map[media_id] = media_hndl
    save_media_map(media_map)
    logger().info('Streaming "%s" on %s', source, media_url)
    return media_hndl



from marshmallow import Schema, pre_dump
from marshmallow.fields import Boolean, Integer, Date, Float, String


class TelegramMessageSchema(Schema):
    """
    Schema for Telegram messages.
    """

    message_id = Integer(required=True)
    chat_id = Integer(required=True)
    creation_date = Date()
    chat_username = String()
    chat_firstname = String()
    chat_lastname = String()
    from_user_id = Integer()
    from_username = String()
    from_firstname = String()
    from_lastname = String()
    text = String()
    caption = String()
    edit_date = Date()
    forward_from_message_id = Integer()
    forward_date = Date()
    photo_file_id = String()
    photo_file_size = Integer()
    photo_width = Integer()
    photo_height = Integer()
    document_file_id = String()
    document_file_name = String()
    document_file_size = Integer()
    document_mime_type = String()
    audio_file_id = String()
    audio_file_size = Integer()
    audio_mime_type = String()
    audio_performer = String()
    audio_title = String()
    audio_duration = Integer()
    location_latitude = Float()
    location_longitude = Float()
    contact_phone_number = String()
    contact_first_name = String()
    contact_last_name = String()
    contact_user_id = Integer()
    contact_vcard = String()
    video_file_id = String()
    video_file_size = Integer()
    video_width = Integer()
    video_height = Integer()
    video_mime_type = String()
    video_duration = Integer()
    link = String()
    media_group_id = String()

    @pre_dump
    def pre_dump(self, msg, **_) -> dict:
        ret = {
            'message_id': msg.message_id,
            'chat_id': msg.chat_id,
            'chat_username': msg.chat.username,
            'chat_firstname': msg.chat.first_name,
            'chat_lastname': msg.chat.last_name,
            'text': msg.text,
            'caption': msg.caption,
            'creation_date': msg.date,
            'edit_date': msg.edit_date,
            'forward_date': msg.forward_date,
            'forward_from_message_id': msg.forward_from_message_id,
            'link': msg.link,
            'media_group_id': msg.media_group_id,
        }

        if msg.from_user:
            ret.update(
                {
                    'from_user_id': msg.from_user.id,
                    'from_username': msg.from_user.username,
                    'from_firstname': msg.from_user.first_name,
                    'from_lastname': msg.from_user.last_name,
                }
            )

        if msg.photo:
            ret.update(
                {
                    'photo_file_id': msg.photo[-1].file_id,
                    'photo_file_size': msg.photo[-1].file_size,
                    'photo_width': msg.photo[-1].width,
                    'photo_height': msg.photo[-1].height,
                }
            )

        if msg.document:
            ret.update(
                {
                    'document_file_id': msg.document.file_id,
                    'document_file_name': msg.document.file_name,
                    'document_file_size': msg.document.file_size,
                    'document_mime_type': msg.document.mime_type,
                }
            )

        if msg.audio:
            ret.update(
                {
                    'audio_file_id': msg.audio.file_id,
                    'audio_file_size': msg.audio.file_size,
                    'audio_mime_type': msg.audio.mime_type,
                    'audio_performer': msg.audio.performer,
                    'audio_title': msg.audio.title,
                    'audio_duration': msg.audio.duration,
                }
            )

        if msg.video:
            ret.update(
                {
                    'video_file_id': msg.video.file_id,
                    'video_file_size': msg.video.file_size,
                    'video_width': msg.video.width,
                    'video_height': msg.video.height,
                    'video_mime_type': msg.video.mime_type,
                    'video_duration': msg.video.duration,
                }
            )

        if msg.location:
            ret.update(
                {
                    'location_latitude': msg.location.latitude,
                    'location_longitude': msg.location.longitude,
                }
            )

        if msg.contact:
            ret.update(
                {
                    'contact_phone_number': msg.contact.phone_number,
                    'contact_first_name': msg.contact.first_name,
                    'contact_last_name': msg.contact.last_name,
                    'contact_user_id': msg.contact.user_id,
                    'contact_vcard': msg.contact.vcard,
                }
            )

        return ret


class TelegramFileSchema(Schema):
    """
    Schema for Telegram files.
    """

    file_id = String(required=True)
    file_path = String(required=True)
    file_size = Integer(required=True)


class TelegramUserSchema(Schema):
    """
    Schema for Telegram users.
    """

    user_id = Integer(required=True, data_key='id')
    username = String(required=True)
    is_bot = Boolean(required=True)
    first_name = String(required=True)
    last_name = String()
    language_code = String()
    link = String()


class TelegramChatSchema(Schema):
    """
    Schema for Telegram chats.
    """

    chat_id = Integer(required=True)
    link = String(required=True)
    username = String(required=True)
    invite_link = String()
    title = String()
    description = String()
    type = String()
    first_name = String()
    last_name = String()



from ._base import BaseMailPlugin
from ._in import MailInPlugin
from ._out import MailOutPlugin
from ._utils import mail_plugins


__all__ = [
    'BaseMailPlugin',
    'MailInPlugin',
    'MailOutPlugin',
    'mail_plugins',
]



import json
from typing import Any, Dict, Optional, Union

from redis import Redis

from platypush.backend import Backend
from platypush.message import Message
from platypush.utils import get_redis_conf


class RedisBackend(Backend):
    """
    Backend that reads messages from a configured Redis queue (default:
    ``platypush/backend/redis``) and forwards them to the application bus.

    Useful when you have plugin whose code is executed in another process
    and can't post events or requests to the application bus.
    """

    def __init__(
        self,
        *args,
        queue: str = 'platypush/backend/redis',
        redis_args: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        """
        :param queue: Name of the Redis queue to listen to (default:
            ``platypush/backend/redis``).

        :param redis_args: Arguments that will be passed to the redis-py
            constructor (e.g. host, port, password). See
            https://redis-py.readthedocs.io/en/latest/connections.html#redis.Redis
            for supported parameters.
        """
        super().__init__(*args, **kwargs)
        self.redis_args = redis_args or get_redis_conf()
        self.queue = queue
        self.redis: Optional[Redis] = None

    def send_message(
        self, msg: Union[str, Message], queue_name: Optional[str] = None, **_
    ):
        """
        Send a message to a Redis queue.

        :param msg: Message to send, as a ``Message`` object or a string.
        :param queue_name: Queue name to send the message to (default:
            configured ``queue`` value).
        """
        if not self.redis:
            self.logger.warning('The Redis backend is not yet running.')
            return

        self.redis.rpush(queue_name or self.queue, str(msg))

    def get_message(self, queue_name: Optional[str] = None) -> Optional[Message]:
        queue = queue_name or self.queue
        assert self.redis, 'The Redis backend is not yet running.'
        data = self.redis.blpop(queue, timeout=1)
        if data is None:
            return None

        msg = data[1].decode()
        try:
            msg = Message.build(msg)
        except Exception as e:
            self.logger.debug(str(e))
            try:
                msg = json.loads(msg)
            except Exception as ee:
                self.logger.exception(ee)

        return msg

    def run(self):
        super().run()
        self.logger.info(
            'Initialized Redis backend on queue %s with arguments %s',
            self.queue,
            self.redis_args,
        )

        with Redis(**self.redis_args) as self.redis:
            while not self.should_stop():
                try:
                    msg = self.get_message()
                    if not msg:
                        continue

                    self.logger.info('Received message on the Redis backend: %s', msg)
                    self.on_message(msg)
                except Exception as e:
                    self.logger.exception(e)

        self.logger.info('Redis backend terminated')


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class PingEvent(Event):
    """Ping event, used for testing purposes"""

    def __init__(self, *args, message=None, **kwargs):
        """
        :param message: Ping message
        :type message: object
        """

        super().__init__(*args, message=message, **kwargs)


class HostDownEvent(Event):
    """
    Event triggered when a remote host stops responding ping requests.
    """

    def __init__(self, host: str, *args, **kwargs):
        super().__init__(host=host, *args, **kwargs)


class HostUpEvent(Event):
    """
    Event triggered when a remote host starts responding ping requests.
    """

    def __init__(self, host: str, *args, **kwargs):
        super().__init__(host=host, *args, **kwargs)


class PingResponseEvent(Event):
    """
    Event triggered when a ping response is received.
    """

    def __init__(
        self,
        host: str,
        min: float,
        max: float,
        avg: float,
        mdev: float,
        *args,
        **kwargs
    ):
        """
        :param host: Remote host IP or name.
        :param min: Minimum round-trip time (in ms).
        :param max: Maximum round-trip time (in ms).
        :param avg: Average round-trip time (in ms).
        :param mdev: Standard deviation of the round-trip time (in ms).
        """
        super().__init__(
            host=host, min=min, max=max, avg=avg, mdev=mdev, *args, **kwargs
        )


# vim:sw=4:ts=4:et:



import asyncio
import importlib
import logging
import os

from dataclasses import dataclass, field
from threading import RLock
from typing import Optional, Any

from ..bus import Bus
from ..config import Config
from ..utils import get_enabled_plugins, get_plugin_name_by_class

logger = logging.getLogger('platypush:context')


@dataclass
class Context:
    """
    Data class to hold the context of the application.
    """

    # backend_name -> backend_instance
    backends: dict = field(default_factory=dict)
    # plugin_name -> plugin_instance
    plugins: dict = field(default_factory=dict)
    # Reference to the main application bus
    bus: Optional[Bus] = None


_ctx = Context()

# # Map: backend_name -> backend_instance
# backends = {}

# # Map: plugin_name -> plugin_instance
# plugins = {}

# Map: plugin_name -> init_lock to make sure that a plugin isn't initialized
# multiple times
plugins_init_locks = {}

# Reference to the main application bus
# main_bus = None


def get_context() -> Context:
    """
    Get the current application context.
    """

    return _ctx


def register_backends(bus=None, global_scope=False, **kwargs):
    """
    Initialize the backend objects based on the configuration and returns a
        name -> backend_instance map.

    Params:
        bus -- If specific (it usually should), the messages processed by the
            backends will be posted on this bus.
        kwargs -- Any additional key-value parameters required to initialize
            the backends
    """

    if bus:
        _ctx.bus = bus

    if global_scope:
        backends = _ctx.backends
    else:
        backends = {}

    for name, cfg in Config.get_backends().items():
        module = importlib.import_module('platypush.backend.' + name)

        # e.g. backend.http main class: HttpBackend
        cls_name = ''
        for token in module.__name__.title().split('.')[2:]:
            cls_name += token.title()
        cls_name += 'Backend'

        try:
            b = getattr(module, cls_name)(bus=bus, **cfg, **kwargs)
            backends[name] = b
        except AttributeError as e:
            logger.warning('No such class in %s: %s', module.__name__, cls_name)
            raise RuntimeError(e) from e

    return backends


def register_plugins(bus=None):
    """
    Register and start all the ``RunnablePlugin`` configured implementations.
    """
    from ..plugins import RunnablePlugin

    for plugin in get_enabled_plugins().values():
        if isinstance(plugin, RunnablePlugin):
            plugin.bus = bus
            plugin.start()


def get_backend(name):
    """Returns the backend instance identified by name if it exists"""

    return _ctx.backends.get(name)


# pylint: disable=too-many-branches
def get_plugin(plugin, plugin_name=None, reload=False):
    """
    Registers a plugin instance by name if not registered already, or returns
    the registered plugin instance.

    :param plugin: Plugin name or class type.
    :param plugin_name: Plugin name, kept only for backwards compatibility.
    :param reload: If ``True``, the plugin will be reloaded if it's already
        been registered.
    """
    from ..plugins import Plugin, RunnablePlugin

    if isinstance(plugin, str):
        name = plugin
    elif plugin_name:
        name = plugin_name
    elif issubclass(plugin, Plugin):
        name = get_plugin_name_by_class(plugin)  # type: ignore
    else:
        raise TypeError(f'Invalid plugin type/name: {plugin}')

    assert name, 'No plugin name provided'
    if name not in plugins_init_locks:
        plugins_init_locks[name] = RLock()

    if name in _ctx.plugins and not reload:
        return _ctx.plugins[name]

    module_name = None
    if isinstance(plugin, str):
        module_name = f'platypush.plugins.{name}'
    elif issubclass(plugin, Plugin):
        module_name = plugin.__module__
    else:
        raise RuntimeError(f'Invalid plugin type/name: {plugin}')

    try:
        plugin = importlib.import_module(module_name)
    except ImportError as e:
        logger.warning('No such plugin: %s', name)
        raise RuntimeError(e) from e

    # e.g. plugins.music.mpd main class: MusicMpdPlugin
    cls_name = ''
    for token in name.split('.'):
        if token:
            cls_name += token[0].upper() + token[1:]
    cls_name += 'Plugin'

    plugin_conf = Config.get_plugins()[name] if name in Config.get_plugins() else {}

    if 'disabled' in plugin_conf:
        if plugin_conf['disabled'] is True:
            return None
        del plugin_conf['disabled']

    if 'enabled' in plugin_conf:
        if plugin_conf['enabled'] is False:
            return None
        del plugin_conf['enabled']

    try:
        plugin_class = getattr(plugin, cls_name)
    except AttributeError as e:
        logger.warning('No such class in %s: %s [error: %s]', name, cls_name, e)
        raise RuntimeError(e) from e

    with plugins_init_locks[name]:
        plugin = _ctx.plugins.get(name)
        if plugin:
            if not reload:
                return _ctx.plugins[name]
            if isinstance(plugin, RunnablePlugin):
                plugin.stop()

        _ctx.plugins[name] = plugin_class(**plugin_conf)

    return _ctx.plugins[name]


def get_bus() -> Bus:
    """
    Get or register the main application bus.
    """
    from platypush.bus.redis import RedisBus
    from platypush.utils import get_redis_conf

    if _ctx.bus:
        return _ctx.bus

    redis_queue = (
        os.environ.get('PLATYPUSH_REDIS_QUEUE') or RedisBus.DEFAULT_REDIS_QUEUE
    )

    _ctx.bus = RedisBus(
        redis_queue=redis_queue,
        **get_redis_conf(),
    )

    return _ctx.bus


def get_or_create_event_loop() -> asyncio.AbstractEventLoop:
    """
    Get or create a new event loop
    """
    try:
        loop = asyncio.get_event_loop()
    except (DeprecationWarning, RuntimeError):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    return loop


class Variable:
    """
    Utility class to wrap platform variables in your custom scripts.
    Usage:

        .. code-block:: python

            # Pass `persisted=False` to get/set an in-memory variable
            # on the Redis instance (default: the variable is
            # persisted on the internal database)
            var = Variable('myvar')
            value = var.get()
            var.set('new value')

    """

    def __init__(self, name: str, persisted: bool = True):
        self.name = name
        self._persisted = persisted

    def get(self) -> Optional[Any]:
        plugin = get_plugin('variable')
        getter = getattr(plugin, 'get' if self._persisted else 'mget')
        return getter(self.name).output.get(self.name)

    def set(self, value: Any):
        plugin = get_plugin('variable')
        setter = getattr(plugin, 'set' if self._persisted else 'mset')
        setter(**{self.name: value})


# vim:sw=4:ts=4:et:



from collections import defaultdict
from contextlib import contextmanager
from queue import Empty, Queue
from threading import Event, RLock, Thread, current_thread
from typing import (
    Dict,
    Final,
    Generator,
    List,
    Optional,
    Union,
)
import bluetooth

from platypush.entities.bluetooth import BluetoothDevice, BluetoothService
from platypush.message.event.bluetooth import (
    BluetoothConnectionFailedEvent,
    BluetoothDeviceConnectedEvent,
    BluetoothDeviceDisconnectedEvent,
)

from ..._manager import BaseBluetoothManager
from ..._types import RawServiceClass
from .._model import BluetoothDeviceBuilder
from ._connection import BluetoothConnection
from ._service import ServiceDiscoverer
from ._types import ConnectionId


class LegacyManager(BaseBluetoothManager):
    """
    Scanner for Bluetooth non-low-energy devices.
    """

    _service_discovery_timeout: Final[int] = 30

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._connections: Dict[ConnectionId, BluetoothConnection] = {}
        """ Maps (address, port) pairs to Bluetooth connections. """
        self._connection_locks: Dict[ConnectionId, RLock] = defaultdict(RLock)
        """ Maps (address, port) pairs to connection locks. """
        self._service_scanned_devices: Dict[str, bool] = {}
        """ Maps the addresses of the devices whose services have been scanned. """

    def get_device(
        self,
        device: str,
        scan_duration: Optional[int] = None,
        _fail_if_not_cached: bool = False,
    ) -> BluetoothDevice:
        """
        Get/discover a device by its address or name.

        :param device: Device address or name.
        :param scan_duration: Overrides the duration of the scan.
        :param _fail_if_not_cached: Throw an assertion error if the device
            hasn't been cached yet.
        """
        duration = scan_duration or self.poll_interval
        dev = self._cache.get(device)
        if dev:
            return dev  # If it's already cached, just return it.

        assert not _fail_if_not_cached, f'Device "{device}" not found'

        # Otherwise, scan for the device.
        self.logger.info('Scanning for device "%s"...', device)
        self.scan(duration=duration)

        # Run the method again, but this time fail if the device has not been
        # found in the latest scan.
        return self.get_device(device, scan_duration, _fail_if_not_cached=True)

    def _get_matching_services(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
    ) -> List[BluetoothService]:
        """
        Given a device and a port or service UUID, return a list of matching
        services.
        """
        assert port or service_uuid, 'Please specify at least one of port/service_uuid'
        dev = self.get_device(device)
        assert dev, f'Device "{device}" not found'

        matching_services = []
        if port:
            matching_services = [srv for srv in dev.services if srv.port == port]
        elif service_uuid:
            uuid = BluetoothService.to_uuid(service_uuid)
            matching_services = [srv for srv in dev.services if uuid == srv.uuid]

        return matching_services

    def _connect_thread(
        self,
        conn_queue: Queue[BluetoothConnection],
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
    ):
        """
        Connection thread that asynchronously pushes a
        :class:`BluetoothConnection` object to a queue when connected, so the
        caller can wait for the connection to complete and handle timeouts.
        """
        dev = self.get_device(device)
        matching_services = self._get_matching_services(device, port, service_uuid)
        assert matching_services, (
            f'No services found on {dev.address} for '
            f'UUID={service_uuid} port={port}'
        )

        conn = BluetoothConnection(
            address=dev.address,
            service=matching_services[0],
            thread=current_thread(),
        )

        existing_conn = self._connections.get(conn.key)
        if existing_conn and existing_conn.socket:
            conn = existing_conn
        else:
            with self._connection_locks[conn.key]:
                addr = conn.address
                port_ = conn.service.port
                self.logger.info(
                    'Opening connection to device %s on port %s', addr, port_
                )

                # Connect to the specified address and port.
                conn.socket.connect((addr, port_))
                self.logger.info('Connected to device %s on port %s', addr, port_)
                self._connections[conn.key] = conn

        conn_queue.put_nowait(conn)

    @contextmanager
    def _connect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
        timeout: Optional[float] = None,
    ) -> Generator[BluetoothConnection, None, None]:
        """
        Wraps the connection thread in a context manager with timeout support.
        """
        dev = self.get_device(device)
        # Queue where the connection object is pushed once the socket is ready
        conn_queue: Queue[BluetoothConnection] = Queue()

        # Start the connection thread
        conn_thread = Thread(
            target=self._connect_thread,
            name=f'Bluetooth:connect@{device}',
            args=(conn_queue, device, port, service_uuid),
        )

        conn_thread.start()

        # Wait for the connection object
        timeout = timeout or self._connect_timeout
        try:
            conn = conn_queue.get(timeout=timeout)
        except Empty as e:
            dev.connected = False
            self.notify(BluetoothConnectionFailedEvent, dev, reason=str(e))
            raise AssertionError(f'Connection to {device} timed out') from e

        dev.connected = True
        conn.service.connected = True
        self.notify(BluetoothDeviceConnectedEvent, dev)
        yield conn

        # Close the connection once the context is over
        with self._connection_locks[conn.key]:
            try:
                conn.close()
            except Exception as e:
                self.logger.warning(
                    'Error while closing the connection to %s: %s', device, e
                )

            self._connections.pop(conn.key, None)

        dev.connected = False
        conn.service.connected = False
        self.notify(BluetoothDeviceDisconnectedEvent, dev)

    def connect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
        interface: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        connected = Event()

        def connect_thread():
            with self._connect(device, port, service_uuid) as conn:
                connected.set()
                conn.stop_event.wait()

            self.logger.info('Connection to %s successfully terminated', conn.address)

        # Start the connection thread
        Thread(
            target=connect_thread,
            name=f'Bluetooth:connect:wrapper@{device}',
        ).start()

        # Wait for the connected event
        timeout = timeout or self._connect_timeout
        conn_success = connected.wait(timeout=timeout)
        assert conn_success, f'Connection to {device} timed out'

    def disconnect(
        self,
        device: str,
        port: Optional[int] = None,
        service_uuid: Optional[RawServiceClass] = None,
    ):
        matching_connections = [
            conn
            for conn in self._connections.values()
            if conn.address == device
            and (port is None or conn.service.port == port)
            and (service_uuid is None or conn.service.uuid == service_uuid)
        ]

        assert matching_connections, f'No active connections found to {device}'
        for conn in matching_connections:
            conn.close()

    def scan(self, duration: Optional[float] = None) -> List[BluetoothDevice]:
        duration = duration or self.poll_interval
        assert duration, 'Scan duration must be set'
        duration = int(max(duration, 1))

        with self._scan_lock:
            # Discover all devices.
            try:
                info = bluetooth.discover_devices(
                    duration=duration, lookup_names=True, lookup_class=True
                )
            except IOError as e:
                self.logger.warning('Could not discover devices: %s', e)
                # Wait a bit before a potential retry
                self.wait_stop(timeout=1)
                return []

        # Pre-fill the services for the devices that have already been scanned.
        services: Dict[str, List[BluetoothService]] = {
            addr: self._cache.get(addr).services  # type: ignore
            for addr, _, __ in info
            if self._cache.get(addr) is not None
        }

        # Check if there are any devices that have not been scanned yet.
        unknown_devices = [
            addr
            for addr, _, __ in info
            if not self._service_scanned_devices.get(addr, False)
        ]

        # Discover the services for the devices that have not been scanned.
        if unknown_devices:
            services.update(
                ServiceDiscoverer().discover(
                    *unknown_devices, timeout=self._service_discovery_timeout
                )
            )

        # Initialize the BluetoothDevice objects.
        devices = {
            addr: BluetoothDeviceBuilder.build(
                address=addr,
                name=name,
                raw_class=class_,
                services=services.get(addr, []),
            )
            for addr, name, class_ in info
        }

        for dev in devices.values():
            self._service_scanned_devices[dev.address] = True
            if self._blacklist.matches(dev):
                self.logger.debug('Ignoring blacklisted device: %s', dev.address)
            else:
                self._device_queue.put_nowait(dev)

        return list(devices.values())

    def read(
        self,
        device: str,
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
        size: int = 1024,
    ) -> bytearray:
        """
        :param size: Number of bytes to read.
        """
        with self._connect(
            device, service_uuid=service_uuid, timeout=connect_timeout
        ) as conn:
            try:
                return conn.socket.recv(size)
            except bluetooth.BluetoothError as e:
                raise AssertionError(f'Error reading from {device}: {e}') from e

    def write(
        self,
        device: str,
        data: Union[bytes, bytearray],
        service_uuid: RawServiceClass,
        interface: Optional[str] = None,
        connect_timeout: Optional[float] = None,
    ):
        with self._connect(
            device, service_uuid=service_uuid, timeout=connect_timeout
        ) as conn:
            try:
                return conn.socket.send(data)
            except bluetooth.BluetoothError as e:
                raise AssertionError(f'Error reading from {device}: {e}') from e

    def run(self):
        super().run()
        self.logger.info('Starting legacy Bluetooth scanner')

        while not self.should_stop():
            scan_enabled = self._scan_enabled.wait(timeout=1)
            if scan_enabled:
                self.scan(duration=self.poll_interval)

    def stop(self):
        super().stop()

        # Close any active connections
        for conn in list(self._connections.values()):
            conn.close(timeout=5)

        self.logger.info('Stopped the Bluetooth legacy scanner')
